import os
import re
import math
from harnice.utils import appearance


def add_entire_svg_file_contents_to_group(filepath, new_group_name):
    """
    Wraps the entire contents of an SVG file in a new group element.

    Reads an SVG file, extracts its inner content (everything between `<svg>` tags),
    and wraps it in a new group element with start and end markers. The original
    file is modified in place.

    **Args:**
    - `filepath` (str): Path to the SVG file to modify.
    - `new_group_name` (str): Name to use for the new group element (will create
        `{new_group_name}-contents-start` and `{new_group_name}-contents-end` markers).

    **Raises:**
    - `ValueError`: If the file does not appear to be a valid SVG or has no inner contents.
    """
    if os.path.exists(filepath):
        try:
            with open(filepath, "r", encoding="utf-8") as file:
                svg_content = file.read()

            match = re.search(r"<svg[^>]*>(.*?)</svg>", svg_content, re.DOTALL)
            if not match:
                raise ValueError(
                    "File does not appear to be a valid SVG or has no inner contents."
                )
            inner_content = match.group(1).strip()

            updated_svg_content = (
                f'<svg xmlns="http://www.w3.org/2000/svg">\n'
                f'  <g id="{new_group_name}-contents-start">\n'
                f"    {inner_content}\n"
                f"  </g>\n"
                f'  <g id="{new_group_name}-contents-end"></g>\n'
                f"</svg>\n"
            )

            with open(filepath, "w", encoding="utf-8") as file:
                file.write(updated_svg_content)

        except Exception as e:
            print(
                f"Error adding contents of {os.path.basename(filepath)} to a new group {new_group_name}: {e}"
            )
    else:
        print(
            f"Trying to add contents of {os.path.basename(filepath)} to a new group but file does not exist."
        )


def find_and_replace_svg_group(
    source_svg_filepath,
    source_group_name,
    destination_svg_filepath,
    destination_group_name,
):
    """
    Copies SVG group content from one file to another, replacing existing group content.

    Extracts the content between group markers in a source SVG file and replaces
    the content between corresponding markers in a destination SVG file. The group
    markers are identified by `{group_name}-contents-start` and `{group_name}-contents-end` IDs.

    **Args:**
    - `source_svg_filepath` (str): Path to the source SVG file containing the group to copy.
    - `source_group_name` (str): Name of the source group to extract content from.
    - `destination_svg_filepath` (str): Path to the destination SVG file to modify.
    - `destination_group_name` (str): Name of the destination group to replace content in.

    **Returns:**
    - `int`: Always returns `1` (success indicator).

    **Raises:**
    - `ValueError`: If any of the required group markers are not found in the source
        or destination files.
    """
    with open(source_svg_filepath, "r", encoding="utf-8") as source_file:
        source_svg_content = source_file.read()
    with open(destination_svg_filepath, "r", encoding="utf-8") as target_file:
        target_svg_content = target_file.read()

    source_start_tag = f'id="{source_group_name}-contents-start"'
    source_end_tag = f'id="{source_group_name}-contents-end"'
    dest_start_tag = f'id="{destination_group_name}-contents-start"'
    dest_end_tag = f'id="{destination_group_name}-contents-end"'

    source_start_index = source_svg_content.find(source_start_tag)
    if source_start_index == -1:
        raise ValueError(
            f"[ERROR] Source start tag <{source_start_tag}> not found in <{source_svg_filepath}>."
        )
    source_start_index = source_svg_content.find(">", source_start_index) + 1

    source_end_index = source_svg_content.find(source_end_tag)
    if source_end_index == -1:
        raise ValueError(
            f"[ERROR] Source end tag <{source_end_tag}> not found in <{source_svg_filepath}>."
        )
    source_end_index = source_svg_content.rfind("<", 0, source_end_index)

    dest_start_index = target_svg_content.find(dest_start_tag)
    if dest_start_index == -1:
        raise ValueError(
            f"[ERROR] Target start tag <{dest_start_tag}> not found in <{destination_svg_filepath}>."
        )
    dest_start_index = target_svg_content.find(">", dest_start_index) + 1

    dest_end_index = target_svg_content.find(dest_end_tag)
    if dest_end_index == -1:
        raise ValueError(
            f"[ERROR] Target end tag <{dest_end_tag}> not found in <{destination_svg_filepath}>."
        )
    dest_end_index = target_svg_content.rfind("<", 0, dest_end_index)

    replacement_group_content = source_svg_content[source_start_index:source_end_index]

    updated_svg_content = (
        target_svg_content[:dest_start_index]
        + replacement_group_content
        + target_svg_content[dest_end_index:]
    )

    with open(destination_svg_filepath, "w", encoding="utf-8") as updated_file:
        updated_file.write(updated_svg_content)

    return 1


def draw_styled_path(spline_points, stroke_width_inches, appearance_dict, local_group):
    """
    Adds a styled spline path to the local group.
    Call as if you were appending any other element to an svg group.

    Spline points are a list of dictionaries with x and y coordinates. [{"x": 0, "y": 0, "tangent": 0}, {"x": 1, "y": 1, "tangent": 0}]
    Appearance dictionary is a dictionary with the following keys: base_color, outline_color, parallelstripe, perpstripe, slash_lines
    Slash lines dictionary is a dictionary with the following keys: direction, angle, step, color, slash_width_inches

    If no appearance dictionary is provided, a rainbow spline will be drawn in place of the path.
    """

    if not appearance_dict:
        appearance_dict = appearance.parse(
            "{'base_color':'red', 'perpstripe':['orange','yellow','green','blue','purple']}"
        )
        stroke_width_inches = 0.01
    else:
        appearance_dict = appearance.parse(appearance_dict)

    # ---------------------------------------------------------------------
    # --- spline_to_path
    # ---------------------------------------------------------------------
    def spline_to_path(points):
        if len(points) < 2:
            return ""
        path = f"M {points[0]['x']:.3f},{-points[0]['y']:.3f}"
        for i in range(len(points) - 1):
            p0, p1 = points[i], points[i + 1]
            t0, t1 = math.radians(p0.get("tangent", 0)), math.radians(
                p1.get("tangent", 0)
            )
            d = math.hypot(p1["x"] - p0["x"], p1["y"] - p0["y"])
            ctrl_dist = d * 0.5
            c1x = p0["x"] + math.cos(t0) * ctrl_dist
            c1y = p0["y"] + math.sin(t0) * ctrl_dist
            c2x = p1["x"] - math.cos(t1) * ctrl_dist
            c2y = p1["y"] - math.sin(t1) * ctrl_dist
            path += f" C {c1x:.3f},{-c1y:.3f} {c2x:.3f},{-c2y:.3f} {p1['x']:.3f},{-p1['y']:.3f}"
        return path

    # ---------------------------------------------------------------------
    # --- draw consistent slanted hatches (twisted wire)
    # ---------------------------------------------------------------------
    def draw_slash_lines(points, slash_lines_dict):
        if slash_lines_dict.get("direction") in ("RH", "LH"):
            if slash_lines_dict.get("angle") is not None:
                angle_slant = slash_lines_dict.get("angle")
            else:
                angle_slant = 20
            if slash_lines_dict.get("step") is not None:
                step_dist = slash_lines_dict.get("step")
            else:
                step_dist = stroke_width * 3
            if slash_lines_dict.get("color") is not None:
                color_line = slash_lines_dict.get("color")
            else:
                color_line = "black"
            if slash_lines_dict.get("color") is not None:
                color_line = slash_lines_dict.get("color")
            else:
                color_line = "black"
            if slash_lines_dict.get("slash_width_inches") is not None:
                slash_width = slash_lines_dict.get("slash_width_inches") * 96
            else:
                slash_width = 0.25

        line_elements = []

        def bezier_eval(p0, c1, c2, p1, t):
            mt = 1 - t
            x = (
                (mt**3) * p0[0]
                + 3 * (mt**2) * t * c1[0]
                + 3 * mt * (t**2) * c2[0]
                + (t**3) * p1[0]
            )
            y = (
                (mt**3) * p0[1]
                + 3 * (mt**2) * t * c1[1]
                + 3 * mt * (t**2) * c2[1]
                + (t**3) * p1[1]
            )
            dx = (
                3 * (mt**2) * (c1[0] - p0[0])
                + 6 * mt * t * (c2[0] - c1[0])
                + 3 * (t**2) * (p1[0] - c2[0])
            )
            dy = (
                3 * (mt**2) * (c1[1] - p0[1])
                + 6 * mt * t * (c2[1] - c1[1])
                + 3 * (t**2) * (p1[1] - c2[1])
            )
            return {"x": x, "y": y, "tangent": math.degrees(math.atan2(dy, dx))}

        def bezier_length(p0, c1, c2, p1, samples=80):
            prev = bezier_eval(p0, c1, c2, p1, 0)
            L = 0.0
            for i in range(1, samples + 1):
                t = i / samples
                pt = bezier_eval(p0, c1, c2, p1, t)
                L += math.hypot(pt["x"] - prev["x"], pt["y"] - prev["y"])
                prev = pt
            return L

        # -------------------------------------------------------------
        # Iterate through Bézier segments
        for i in range(len(points) - 1):
            p0 = (points[i]["x"], points[i]["y"])
            p1 = (points[i + 1]["x"], points[i + 1]["y"])
            t0 = math.radians(points[i].get("tangent", 0))
            t1 = math.radians(points[i + 1].get("tangent", 0))
            d = math.hypot(p1[0] - p0[0], p1[1] - p0[1])
            ctrl_dist = d * 0.5
            c1 = (p0[0] + math.cos(t0) * ctrl_dist, p0[1] + math.sin(t0) * ctrl_dist)
            c2 = (p1[0] - math.cos(t1) * ctrl_dist, p1[1] - math.sin(t1) * ctrl_dist)

            L = bezier_length(p0, c1, c2, p1)
            num_steps = max(1, int(L / step_dist))
            step_dist_actual = L / num_steps  # uniform spacing

            for z in range(num_steps + 1):
                # ------------------------------------------------------------------
                # 1. Evaluate point along Bézier curve
                # ------------------------------------------------------------------
                t_norm = min(1.0, (z * step_dist_actual) / L)
                P = bezier_eval(p0, c1, c2, p1, t_norm)

                # Centerpoint on spline
                center = (P["x"], P["y"])

                # ------------------------------------------------------------------
                # 2. Tangent and hatch angle computation
                # ------------------------------------------------------------------
                # Tangent direction of the spline at this point (radians)
                tangent_angle = math.radians(P["tangent"])

                # LH vs RH determines whether we add or subtract the slant
                if slash_lines_dict.get("direction") == "LH":
                    line_angle = tangent_angle + math.radians(angle_slant)
                else:  # "RH"
                    line_angle = tangent_angle - math.radians(angle_slant)

                # ------------------------------------------------------------------
                # 3. Compute hatch line geometry
                # ------------------------------------------------------------------
                # Shorter lines at steep slant; normalize by cos(slant)
                line_length = stroke_width / math.sin(math.radians(angle_slant))

                # Vector along hatch direction
                dx = math.cos(line_angle) * (line_length / 2)
                dy = math.sin(line_angle) * (line_length / 2)

                # Line endpoints
                x1 = center[0] - dx
                y1 = center[1] - dy
                x2 = center[0] + dx
                y2 = center[1] + dy

                # ------------------------------------------------------------------
                # 4. Append SVG element
                # ------------------------------------------------------------------
                line_elements.append(
                    f'<line x1="{x1:.2f}" y1="{-y1:.2f}" '
                    f'x2="{x2:.2f}" y2="{-y2:.2f}" '
                    f'stroke="{color_line}" stroke-width="{slash_width}" />'
                )

        local_group.extend(line_elements)

    # ---------------------------------------------------------------------
    # --- Main body rendering
    # ---------------------------------------------------------------------
    base_color = appearance_dict.get("base_color", "white")
    outline_color = appearance_dict.get("outline_color")
    path_d = spline_to_path(spline_points)

    # outline path
    stroke_width = stroke_width_inches * 96
    if outline_color:
        local_group.append(
            f'<path d="{path_d}" stroke="{outline_color}" stroke-width="{stroke_width}" '
            f'fill="none" stroke-linecap="round" stroke-linejoin="round"/>'
        )
        stroke_width = stroke_width - 0.5

    # base path
    local_group.append(
        f'<path d="{path_d}" stroke="{base_color}" stroke-width="{stroke_width}" '
        f'fill="none" stroke-linecap="round" stroke-linejoin="round"/>'
    )

    # ---------------------------------------------------------------------
    # --- Add pattern overlays
    # ---------------------------------------------------------------------
    if appearance_dict.get("parallelstripe"):
        stripes = appearance_dict["parallelstripe"]
        num = len(stripes)
        stripe_thickness = stroke_width / num
        stripe_spacing = stroke_width / num
        offset = -(num - 1) * stripe_spacing / 2
        for color in stripes:
            local_group.append(
                f'<path d="{path_d}" stroke="{color}" '
                f'stroke-width="{stripe_thickness}" fill="none" '
                f'transform="translate(0,{offset:.2f})"/>'
            )
            offset += stripe_spacing

    if appearance_dict.get("perpstripe"):
        stripes = appearance_dict["perpstripe"]
        num = len(stripes)
        pattern_length = 30
        dash = pattern_length / (num + 1)
        gap = pattern_length - dash
        offset = 0
        for color in stripes:
            offset += dash
            local_group.append(
                f'<path d="{path_d}" stroke="{color}" stroke-width="{stroke_width}" '
                f'stroke-dasharray="{dash},{gap}" stroke-dashoffset="{offset}" '
                f'fill="none" />'
            )

    # --- Slash lines ---
    if appearance_dict.get("slash_lines") is not None:
        slash_lines_dict = appearance_dict.get("slash_lines")
        draw_slash_lines(spline_points, slash_lines_dict)


def table(layout_dict, format_dict, columns_list, content_list, path_to_svg, contents_group_name):
    """
    This function is called when the user needs to build a general SVG table.
    ```python
    svg_utils.table(
        layout_dict,
        format_dict,
        columns_list,
        content_list,
        path_to_caller,
        svg_name
    )
    ```
    ### Arguments
    - `layout_dict` expects a dictionary describing in which direction the table is built
    - `format_dict` expects a dictionary containing a description of how you want your table to appear.
    - `columns_list` expects a list containing your column header content, width, and formatting rules.
    - `content_list` expects a list containing what is actually presented on your table.

    ### Returns
    - A string of SVG primatives in xml format intended to look like a table. 

    ---

    ## 1. Layout

    The SVG origin (0,0) must exist somewhere in your table. Defining this correctly will help later when tables dynamically update with changing inputs. 

    *example:*
    ```json
    layout = {
        "origin_corner": "top-left",
        "build_direction": "down",
    }
    ```
    Both fields are required. 
    ### Origin Corner

    The origin is defined to be at one of the four corners of the first row `content[0]`. Valid options:
    - `top-left`
    - `top-right`
    - `bottom-left`
    - `bottom-right`

    ### Build Direction 
    When building a table, you can choose to build rows downwards (below the previous, positive y in svg coords) or upwards (above the previous, negative y in svg coords). The direction property defines this:
    - `down` → rows appear below the previous
    - `up` → new rows appear above the previous

    ---

    ## 2. Format

    The format dictionary defines appearance and style of your table. 

    Any number of appearance keys can be defined and named with an identifier that is called when printing that row. This allows you to have rows that whose appearance can be varied dynamically with the table contents. 

    *example:*
    ```json
    format_dict={
        "globals": {
            "font_size": 11,
            "row_height": 20,
        },
        "header": {
            "font_weight":"B",
            "fill_color": "lightgray",
        },
        "row_with_bubble": {
            "row_height": 40,
        },
    }
    ```
    The only reserved key is `globals` which can optionally be used to define fallbacks for any row that does not have a style explicitly called out.

    ### Format Arguments
    Any of the following keys can be defined in any of the format dictionaries. 
    - `font_size` *(number, default=12)* Default font size (px) for all text
    - `font_family` *(string, default=helvetica)* Default font family (e.g., "Arial", "Helvetica")
    - `font_weight`*(`BIU`, default=None)* Add each character for bold, italic, or underline
    - `row_height` *(number, default=18)* Self-explanatory (px)
    - `padding` *(number, default=3)* Default text inset from border if not center or middle justified
    - `line_spacing` *(number, default=14)* Vertical spacing between multi-line text entries
    - `justify` *(`left` \ `center` \ `right`, default=center)* Default horizontal alignment
    - `valign` *(`top` \ `middle` \ `bottom`, default=center)* Default vertical alignment
    - `fill_color` *(default=white)* Cell background color
    - `stroke_color` *(default=black)* Border line color
    - `stroke_width` *(number, default=1)* Border width
    - `text_color` *(default=black)* Default text color

    ### Style Resolution Order
    If something is defined at the row level, it takes precedent over some parameter defined at the column level, which takes precedent over a definition in key `globals`, if defined. If something is not defined at all, the above defaults will apply. 

    ### Color Standard
    - Default color: **black**
    - Accepted formats:
    - Named SVG colors https://www.w3.org/TR/SVG11/types.html#ColorKeywords
    - Hex values (#RGB or #RRGGBB)

    ---

    ## 3. Columns

    The column argument is a list of dictionaries containing definition of how many columns there are, the order in which they exist, how to reference them, and any applicable formatting.

    *ex:*
    ```json
    columns_list=[
        {
            "name": "rev"
            "width": 60,
            "justify": "center"
        },
        {
            "name": "updated"
            "width": 260,
        },
            "name": "status"
            "width": 120,
            "fill_color": "yellow",
        }
    ]
    ```

    ### Column Arguments
    Each field must have the following required keys:
    - `name` *(string)* Used to identify a column when defining contents later. Must be unique.
    - `width` *(number)* Self-explanatory (px)

    You may add any formatting key as defined in the formatting section as needed.

    Note that the order of the items in the list represents the order in which they will be printed from left to right, regardless of the layout you've chosen for this table.

    ---

    ## 4. Content Structure

    The table content will be referenced from information stored in this argument. It is a list (rows) of dictionaries (columns).

    ```json
    content_list = [
        {
            "format_key": "header"
            "columns": {
                "rev": "REV",
                "updated": "UPDATED",
                "status": "STATUS",
            }
        },
        {
            "columns": {
                "rev": "1",
                "updated": "12/6/25",
                "status": "requires review",
            }
        },
        {
            "columns": {
                "rev": "2",
                "updated": ["12/6/25", "update incomplete"],
            }
        },
        {
            "format_key": "row_with_bubble",
            "columns": {
                "rev": {
                    "instance_name": "rev3-bubble",
                    "item_type": "flagnote"
                },
                "updated": "12/6/25",
                "status": "clear"
            }
        }
    ]
    ```

    Content (the root argument) is a list. Each entry of the root list is representative of a row's worth of data.

    Each entry of that list must contain the dictionary `columns` and may contain dictionary `format_key`. 

    `format_key` may only contain one value which corresponds to the name of a key in the format dictionary. It represents the appearance of that row. If it is not defined, the format of that row will fall back to globals and defaults. Custom formatting of individual cells is not supported. 

    `columns` is a dictionary that contains the actual content you want to appear in each column. The name of each key at this level must match one of the keys in the `columns` argument. It is agnostic to order, and by leaving a key out, simply nothing will appear in that cell. Existing formatting (cell fill and border) will still apply.

    The value of each column key may take one of the following forms:
    - string or number → single-line text, prints directly
    - list[str] → multi-line text where the 0th element prints highest within the cell. Use format key `line_spacing` as needed. 
    - dict → custom 

    ### Importing a Symbol into a Cell

    If you add a dictionary to one of the content cells, content start/end groups will be written into your svg. This will allow the user to generate and/or import symbols into the table using their own logic, without regard for placement into the table.

    ```python
    #from your macro or wherever you're building the table from...
    example_symbol = {
        "lib_repo": instance.get("lib_repo"),
        "item_type": "flagnote",
        "mpn": instance.get("mpn"),
        "instance_name": f"bubble{build_note_number}",
        "note_text": build_note_number,
    }
    symbols_to_build=[example_symbol]
            
    svg_utils.table(
        layout_dict,
        format_dict,
        columns_list,
        content_list,
        os.dirname(path_to_table_svg),
        artifact_id
    )

    # user import logic 
    for symbol in symbols_to_build:
        path_to_symbol = #...
        library_utils.pull(
            symbol,
            update_instances_list=False,
            destination_directory=path_to_symbol,
        )

        svg_utils.find_and_replace_svg_group(
            os.path.join(path_to_symbol, f"{symbol.get('instance_name')}-drawing.svg"),
            symbol.get("instance_name"),
            path_to_table_svg,
            symbol.get("instance_name")
        )
    """

    # --- Default Style Values (Local to the table() function) ---
    DEFAULTS = {
        "font_size": 12,
        "font_family": "helvetica",
        "font_weight": None,
        "row_height": 0.16 * 96,
        "padding": 3,
        "line_spacing": 14,
        "justify": "center",
        "valign": "middle",
        "fill_color": "white",
        "stroke_color": "black",
        "stroke_width": 1,
        "text_color": "black",
    }

    # --- Helper functions ---

    def _validate_layout(layout):
        """Validates the layout dictionary."""
        valid_corners = {"top-left", "top-right", "bottom-left", "bottom-right"}
        valid_directions = {"down", "up"}
        corner = layout.get("origin_corner")
        direction = layout.get("build_direction")
        if corner not in valid_corners:
            raise ValueError(f"Invalid origin_corner: {corner}. Must be one of {valid_corners}.")
        if direction not in valid_directions:
            raise ValueError(f"Invalid build_direction: {direction}. Must be one of {valid_directions}.")
        return layout

    def _validate_columns(columns_list):
        """Validates the columns list, ensuring 'name' and 'width' are present."""
        if not columns_list:
            raise ValueError("columns_list cannot be empty.")
        names = set()
        for col in columns_list:
            if 'name' not in col or 'width' not in col:
                raise ValueError("Each column must have 'name' and 'width' keys.")
            if col['name'] in names:
                raise ValueError(f"Duplicate column name: {col['name']}.")
            names.add(col['name'])
        return columns_list

    def _resolve_style(row_format_key, col_data, format_dict, columns_list):
        """Resolves the final style for a cell."""
        style = DEFAULTS.copy()
        style.update(format_dict.get("globals", {}))
        col_name = col_data.get('name')
        column_def = next((c for c in columns_list if c['name'] == col_name), {})
        style.update({k: v for k, v in column_def.items() if k in DEFAULTS})
        row_format = format_dict.get(row_format_key)
        if row_format:
            style.update({k: v for k, v in row_format.items() if k in DEFAULTS})
        return style

    def _generate_symbol_placeholder(instance_name, bubble_x, bubble_y):
        """Generates the SVG group structure placeholder."""
        svg_lines = []
        svg_lines.append(f'<g id="{instance_name}" transform="translate({bubble_x},{bubble_y})">')
        svg_lines.append(f'  <g id="{instance_name}-contents-start">')
        svg_lines.append("  </g>")
        svg_lines.append(f'  <g id="{instance_name}-contents-end"/>')
        svg_lines.append("</g>")
        return "\n".join(svg_lines)

    def _draw_cell_content(x, y, width, height, content, style, columns_list, format_dict):
        """Generates the SVG for the cell's content (text or symbol)."""
        svg_primitives = []
        instances_to_copy_in = []
        
        if isinstance(content, dict) and 'instance_name' in content:
            instance_name = content['instance_name']
            item_type = content.get('item_type') 
            
            align_x = width / 2 if style['justify'] == 'center' else (width - style['padding'] if style['justify'] == 'right' else style['padding'])
            align_y = height / 2 if style['valign'] == 'middle' else (height - style['padding'] if style['valign'] == 'bottom' else style['padding'])

            symbol_xml = _generate_symbol_placeholder(instance_name=instance_name, bubble_x=x + align_x, bubble_y=y + align_y)
            instances_to_copy_in.append({"item_type":item_type, "instance_name":instance_name})
            svg_primitives.append(symbol_xml)

        else:
            text_lines = [str(content)] if isinstance(content, (str, int, float)) else [str(line) for line in content] if isinstance(content, list) else []
            if not text_lines:
                return "", []

            x_pos = x
            text_anchor = "start"
            if style['justify'] == 'center': 
                x_pos += width / 2
                text_anchor = "middle"
            elif style['justify'] == 'right': 
                x_pos += width - style['padding']
                text_anchor = "end"
            elif style['justify'] == 'left': 
                x_pos += style['padding']
                text_anchor = "start"

            num_lines = len(text_lines)
            total_text_height = (num_lines - 1) * style['line_spacing'] + style['font_size']
            y_start = y
            if style['valign'] == 'middle': 
                y_start += (height - total_text_height) / 2
            elif style['valign'] == 'bottom': 
                y_start += height - total_text_height - style['padding']
            elif style['valign'] == 'top': 
                y_start += style['padding']
            y_pos = y_start + style['font_size'] * 0.8 

            text_style = {"font-size": f"{style['font_size']}px", "font-family": style['font_family'], "fill": style['text_color'], "text-anchor": text_anchor}
            if style['font_weight']:
                if 'B' in style['font_weight']: 
                    text_style['font-weight'] = 'bold'
                if 'I' in style['font_weight']: 
                    text_style['font-style'] = 'italic'
                if 'U' in style['font_weight']: 
                    text_style['text-decoration'] = 'underline'
            style_str = "; ".join(f"{k}: {v}" for k, v in text_style.items())

            for i, line in enumerate(text_lines):
                current_y = y_pos + i * style['line_spacing']
                svg_primitives.append(f'<text x="{x_pos}" y="{current_y}" style="{style_str}">{line}</text>')

        return "\n".join(svg_primitives), instances_to_copy_in

    def _draw_cell_rect(x, y, width, height, style):
        """Generates the SVG for the cell's background and border."""
        rect_style = {"fill": style['fill_color'], "stroke": style['stroke_color'], "stroke-width": style['stroke_width']}
        style_str = "; ".join(f"{k}: {v}" for k, v in rect_style.items())
        return f'<rect x="{x}" y="{y}" width="{width}" height="{height}" style="{style_str}" />'

    # ====================================================================
    # MAIN FUNCTION LOGIC START
    # ====================================================================

    # 1. INITIAL VALIDATION (and state definition)
    layout = _validate_layout(layout_dict)
    columns = _validate_columns(columns_list)
    total_width = sum(col['width'] for col in columns)

    # ... (Setup for BUILD_SVG LOGIC remains unchanged) ...
    svg_rows = []
    col_x_starts = [0]
    current_x = 0
    for col in columns[:-1]:
        current_x += col['width']
        col_x_starts.append(current_x)

    row_heights = [
        _resolve_style(row.get('format_key'), {}, format_dict, columns).get('row_height', DEFAULTS['row_height'])
        for row in content_list
    ]
    total_table_height = sum(row_heights)
    row_height_0 = row_heights[0]
    
    current_y_offset = 0 
    instances_to_copy_in = []
    
    # 3. BUILD ROW BY ROW

    # reverse content list if building upwards so the first row stays on top but is now printed last
    if layout['build_direction'] == 'up':
        content_list = list(reversed(content_list))

    for row_index, row_data in enumerate(content_list):
        row_key = row_data.get('format_key')
        row_height = row_heights[row_index]
        row_svg = []
        
        if layout['build_direction'] == 'down':
            cell_y_start = current_y_offset
            current_y_offset += row_height
        elif layout['build_direction'] == 'up':
            current_y_offset -= row_height
            cell_y_start = current_y_offset

        for col_index, col_def in enumerate(columns):
            col_name = col_def['name']
            col_width = col_def['width']
            cell_content = row_data.get('columns', {}).get(col_name)
            cell_style = _resolve_style(row_key, col_def, format_dict, columns)
            cell_x_start = col_x_starts[col_index]
            
            rect_svg = _draw_cell_rect(cell_x_start, cell_y_start, col_width, row_height, cell_style)
            row_svg.append(rect_svg)
            
            if cell_content is not None:
                content_svg, instances = _draw_cell_content(
                    cell_x_start, cell_y_start, col_width, row_height, 
                    cell_content, cell_style, columns, format_dict
                )
                row_svg.append(content_svg)
                instances_to_copy_in.extend(instances)

        svg_rows.append("\n".join(row_svg))
        
    # 4. FINAL TRANSFORM CALCULATION
    tx = 0
    ty = 0
    
    if layout['origin_corner'] in ('top-right', 'bottom-right'):
        tx = -total_width
    
    if layout['build_direction'] == 'down':
        if layout['origin_corner'] in ('bottom-left', 'bottom-right'):
            ty = -row_height_0
    elif layout['build_direction'] == 'up':
        if layout['origin_corner'] in ('top-left', 'top-right'):
            ty = total_table_height
        
    table_contents = "\n".join(svg_rows)

    # 5. WRAP, SAVE, AND INJECT
    group_output = f"""<svg xmlns="http://www.w3.org/2000/svg" width="1000" height="1000">
  <g id="{contents_group_name}-contents-start">
    <g id="translate" transform="translate({tx}, {ty})">
{table_contents}
    </g>
  </g>
  <g id="{contents_group_name}-contents-end"/>
</svg>"""

    # Save the master SVG file
    with open(path_to_svg, "w") as f:
        f.write(group_output.strip())



import argparse
import os
import sys
import shutil
from harnice.lists import rev_history
from harnice import state
from harnice import fileio


def print_import_status(
    instance_name, item_type, library_status, import_state, called_from_base_directory
):
    print(
        f"{'':<4}"
        f"{instance_name:<40}"
        f"{item_type:<16}"
        f"{library_status:<16}"
        f"{import_state:<32}"
        f"{called_from_base_directory:<32}"
    )


def print_import_status_headers():
    print_import_status(
        "INSTANCE NAME",
        "ITEM TYPE",
        "LIBRARY STATUS",
        "IMPORT STATE",
        "CALLED FROM BASE DIRECTORY",
    )


def main():
    # Ensure cwd exists
    try:
        cwd = os.getcwd()
    except (FileNotFoundError, PermissionError):
        sys.exit(
            "Error: The current working directory is invalid "
            "(it may have been deleted or you lack permission to access it)."
        )

    if not os.path.exists(cwd):
        sys.exit(f"Error: The current working directory no longer exists: {cwd}")

    # -----------------------------
    # Argument parsing
    # -----------------------------
    parser = argparse.ArgumentParser(
        prog="harnice",
        description="Electrical system CAD",
    )

    group = parser.add_mutually_exclusive_group(required=True)

    group.add_argument(
        "-r", "--render", action="store_true", help="Render the product normally"
    )

    group.add_argument(
        "-l",
        "--lightweight",
        action="store_true",
        help="Render the product quickly without performing all checks",
    )

    group.add_argument(
        "--newrev",
        action="store_true",
        help="Create a new revision in the current working directory",
    )

    group.add_argument(
        "--gui",
        action="store_true",
        help="Launch the Harnice GUI launcher",
    )

    args = parser.parse_args()

    if args.gui:
        from harnice.gui.launcher import main as gui_main

        gui_main()
        return

    # -----------------------------
    # Handle new revision creation and exit
    # -----------------------------
    if args.newrev:
        newrev()
        return

    # -----------------------------
    # Ensure we are inside a revision folder
    # May change cwd if new PN created
    # -----------------------------
    fileio.verify_revision_structure()
    item_type = rev_history.info(field="product")

    # -----------------------------
    # Load product module
    # -----------------------------
    try:
        product_module = __import__(
            f"harnice.products.{item_type}", fromlist=[item_type]
        )
    except ModuleNotFoundError:
        sys.exit(f"Unknown product: '{item_type}'")

    # -----------------------------
    # Set the default fileio structure dict to the product's file_structure()
    # -----------------------------
    if hasattr(product_module, "file_structure"):
        structure = product_module.file_structure()
        state.set_file_structure(structure)
    else:
        sys.exit(f"Product '{item_type}' must define file_structure()")

    # -----------------------------
    # Generate product file structure
    # -----------------------------
    if hasattr(product_module, "generate_structure"):
        product_module.generate_structure()
    else:
        sys.exit(f"Product '{item_type}' must define generate_structure()")

    # -----------------------------
    # Execute render logic
    # -----------------------------
    if args.lightweight:
        try:
            product_module.render(lightweight=True)
        except TypeError:
            sys.exit(f"Product '{item_type}' does not support lightweight rendering")
    else:
        product_module.render()

    return


def prompt(text, default=None):
    p = f"{text}"
    if default:
        p += f" [{default}]"
    p += ": "
    return input(p).strip() or default


def newrev():
    from harnice import fileio

    """
    Create a new revision directory by copying the current revision's contents
    and updating filenames to reflect the new revision number.
    """
    # Ensure revision structure is valid and get context
    fileio.verify_revision_structure()

    # Prompt user for new revision number
    new_rev_number = prompt(
        f"Current rev number: {state.partnumber('R')}. Enter new rev number:",
        default=str(int(state.partnumber("R")) + 1),
    )

    # Construct new revision directory path
    new_rev_dir = os.path.join(
        fileio.part_directory(), f"{state.partnumber('pn')}-rev{new_rev_number}"
    )

    # Ensure target directory does not already exist
    if os.path.exists(new_rev_dir):
        raise FileExistsError(f"Revision directory already exists: {new_rev_dir}")

    shutil.copytree(fileio.rev_directory(), new_rev_dir)

    # Walk the new directory and rename all files containing the old rev number
    for root, _, files in os.walk(new_rev_dir):
        for filename in files:
            new_suffix = f"rev{new_rev_number}"

            if state.partnumber("rev") in filename:
                old_path = os.path.join(root, filename)
                new_name = filename.replace(state.partnumber("rev"), new_suffix)
                new_path = os.path.join(root, new_name)

                os.rename(old_path, new_path)

    print(
        f"Successfully created new revision: {state.partnumber('pn')}-rev{new_rev_number}. Please cd into it."
    )


def select_product_type():
    from pathlib import Path
    import harnice.products as products_pkg
    from prompt_toolkit import prompt
    from prompt_toolkit.completion import WordCompleter

    def get_product_types():
        products_dir = Path(products_pkg.__file__).parent
        return sorted(
            p.stem
            for p in products_dir.glob("*.py")
            if p.name != "__init__.py"
        )

    product_types = get_product_types()
    product_map = {p.lower(): p for p in product_types}

    completer = WordCompleter(
        product_types,
        ignore_case=True,
        sentence=True,
    )

    while True:
        value = prompt(
            "What product type are you working on? ",
            completer=completer,
            default="harness",
        ).strip()

        if not value:
            value = "harness"

        key = value.lower()
        if key in product_map:
            return product_map[key]

        print(
            f"Unrecognized product type '{value}'. "
            f"Valid options: {', '.join(product_types)}"
        )


import os
import csv
from harnice import fileio
from harnice.lists import signals_list
from harnice.products import chtype

COLUMNS = [
    "net", #documentation needed
    "circuit_id", #documentation needed
    "signal", #documentation needed
    "net_from_refdes", #documentation needed
    "net_from_channel_id", #documentation needed
    "net_from_connector_name", #documentation needed
    "net_from_cavity", #documentation needed
    "net_to_refdes", #documentation needed
    "net_to_channel_id", #documentation needed
    "net_to_connector_name", #documentation needed
    "net_to_cavity", #documentation needed
    "from_side_device_refdes", #documentation needed
    "from_side_device_chname", #documentation needed
    "to_side_device_refdes", #documentation needed
    "to_side_device_chname", #documentation needed
    "from_channel_type", #documentation needed
    "to_channel_type", #documentation needed
]


def new():
    """
    Makes a new blank circuits list. Overwrites existing circuits list.
    
    Args: none
    
    Returns: none
    """
    # --- helper: first non-empty field ---
    def first_nonempty(row, *candidate_names):
        for name in candidate_names:
            value = (row.get(name) or "").strip()
            if value:
                return value
        return ""

    # --- load disconnect map and build index ---
    disconnect_index = {}

    for row in fileio.read_tsv("disconnect map"):
        a_refdes = first_nonempty(
            row,
            "A-side_device_refdes",
            "from_destination_device_refdes",
            "from_device_refdes",
        )
        if not a_refdes:
            continue  # skip "available channel" rows

        a_channel_id = first_nonempty(
            row,
            "A-side_device_channel_id",
            "from_destination_device_channel_id",
            "from_device_channel_id",
        )
        b_refdes = first_nonempty(
            row,
            "B-side_device_refdes",
            "to_destination_device_refdes",
            "to_device_refdes",
        )
        b_channel_id = first_nonempty(
            row,
            "B-side_device_channel_id",
            "to_destination_device_channel_id",
            "to_device_channel_id",
        )

        disconnect_refdes = first_nonempty(row, "disconnect_refdes")
        disconnect_channel_id = first_nonempty(
            row, "disconnect_channel_id", "disconnect_channel_id"
        )

        key_forward = (
            a_refdes,
            a_channel_id,
            b_refdes,
            b_channel_id,
            disconnect_refdes,
        )
        key_reverse = (
            b_refdes,
            b_channel_id,
            a_refdes,
            a_channel_id,
            disconnect_refdes,
        )

        disconnect_index[key_forward] = disconnect_channel_id
        disconnect_index[key_reverse] = disconnect_channel_id

    circuits_list = []
    circuit_id = 0

    # --- resolvers ---
    def resolve_device_endpoint(refdes, channel_id, signal):
        device_signals_list_path = os.path.join(
            fileio.dirpath("instance_data"),
            "device",
            refdes,
            f"{refdes}-signals_list.tsv",
        )
        connector_name = (
            signals_list.connector_name_of_channel(channel_id, device_signals_list_path)
            if channel_id
            else ""
        )
        cavity = (
            signals_list.cavity_of_signal(channel_id, signal, device_signals_list_path)
            if channel_id
            else ""
        )
        return {
            "refdes": refdes,
            "channel_id": channel_id,
            "connector_name": connector_name,
            "cavity": cavity,
        }

    def resolve_disconnect_endpoint(refdes, side, signal, channel_id):
        disconnect_signals_list_path = os.path.join(
            fileio.dirpath("instance_data"), "disconnect", refdes, f"{refdes}-signals_list.tsv"
        )

        row = None
        for disconnect_signal_row in fileio.read_tsv(disconnect_signals_list_path):
            if disconnect_signal_row.get("signal", "").strip() == signal.strip():
                if (
                    disconnect_signal_row.get("channel_id", "").strip()
                    == channel_id.strip()
                ):
                    row = disconnect_signal_row
                    break

        if row is None:
            raise ValueError(
                f"Signal '{signal}' of channel '{channel_id}' not found in {disconnect_signals_list_path}"
            )

        cavity = (row.get(f"{side}_cavity") or "").strip()
        return {
            "refdes": refdes,
            "channel_id": channel_id,
            "connector_name": side,
            "cavity": cavity,
        }

    # --- iterate channel map rows ---
    for row in fileio.read_tsv("channel map"):
        if not row.get("from_device_channel_id"):
            continue
        if not row.get("to_device_refdes") and not row.get("multi_ch_junction_id"):
            continue

        from_refdes = row["from_device_refdes"].strip()
        from_channel_id = row["from_device_channel_id"].strip()
        to_refdes = row["to_device_refdes"].strip()
        to_channel_id = row["to_device_channel_id"].strip()

        signals = chtype.signals(row.get("from_channel_type"))

        # --- parse disconnect requirement ---
        disconnect_chain = []
        if row.get("disconnect_refdes_requirement"):
            for token in row["disconnect_refdes_requirement"].split(";"):
                token = token.strip()
                if token:
                    refdes, sides = token.split("(", 1)
                    refdes = refdes.strip()
                    sides = sides.rstrip(")")
                    side_from, side_to = [s.strip() for s in sides.split(",")]
                    disconnect_chain.append((refdes, side_from, side_to))
        disconnect_set = {d[0] for d in disconnect_chain}

        # --- nets list ---
        nets = [n.strip() for n in row.get("chain_of_nets", "").split(";") if n.strip()]

        # --- connection steps (disconnects + final device) ---
        connection_steps = disconnect_chain + [(to_refdes, None, None)]

        if len(connection_steps) != len(nets):
            step_labels = [s[0] for s in connection_steps]
            raise ValueError(
                f"While building circuits from channel_id '{from_channel_id}' of device '{from_refdes}' "
                f"to channel_id '{to_channel_id}' of device '{to_refdes}', "
                f"found {len(connection_steps)} connection steps: "
                f"{', '.join(step_labels) or 'none'}, "
                f"but expected {len(nets)} because there are {len(nets)} nets "
                f"from channel end '{from_channel_id}' to channel end '{to_channel_id}' "
                f"({'; '.join(nets) or 'no nets listed'}). "
                "Each net should correspond to one physical connection segment between devices or disconnects. "
                "Check the channel map for missing or unexpected info in cells, or if the disconnect requirements match the disconnect map."
            )

        # --- iterate signals ---
        for signal in signals:
            current_refdes = from_refdes
            current_side = None
            current_channel_id = from_channel_id

            for step, net in zip(connection_steps, nets):
                refdes, side_from, side_to = step

                if side_from is not None:
                    # disconnect step
                    disconnect_key = (
                        from_refdes,
                        from_channel_id,
                        to_refdes,
                        to_channel_id,
                        refdes,
                    )
                    mapped_channel_id = disconnect_index[disconnect_key]

                    if current_refdes in disconnect_set:
                        left = resolve_disconnect_endpoint(
                            current_refdes, current_side, signal, current_channel_id
                        )
                    else:
                        left = resolve_device_endpoint(
                            current_refdes, current_channel_id, signal
                        )

                    right = resolve_disconnect_endpoint(
                        refdes, side_from, signal, mapped_channel_id
                    )

                    circuits_list.append(
                        {
                            "net": net,
                            "circuit_id": circuit_id,
                            "from_channel_type": row.get("from_channel_type"),
                            "to_channel_type": row.get("to_channel_type"),
                            "signal": signal,
                            "net_from_refdes": left["refdes"],
                            "net_from_channel_id": left["channel_id"],
                            "net_from_connector_name": left["connector_name"],
                            "net_from_cavity": left["cavity"],
                            "net_to_refdes": right["refdes"],
                            "net_to_channel_id": right["channel_id"],
                            "net_to_connector_name": right["connector_name"],
                            "net_to_cavity": right["cavity"],
                            "from_side_device_refdes": from_refdes,
                            "from_side_device_chname": from_channel_id,
                            "to_side_device_refdes": to_refdes,
                            "to_side_device_chname": to_channel_id,
                        }
                    )
                    circuit_id += 1

                    current_refdes = refdes
                    current_side = side_to
                    current_channel_id = mapped_channel_id

                else:
                    # final device step
                    if current_refdes in disconnect_set:
                        left = resolve_disconnect_endpoint(
                            current_refdes, current_side, signal, current_channel_id
                        )
                    else:
                        left = resolve_device_endpoint(
                            current_refdes, current_channel_id, signal
                        )

                    right = resolve_device_endpoint(refdes, to_channel_id, signal)

                    circuits_list.append(
                        {
                            "net": net,
                            "circuit_id": circuit_id,
                            "signal": signal,
                            "from_channel_type": row.get("from_channel_type"),
                            "to_channel_type": row.get("to_channel_type"),
                            "net_from_refdes": left["refdes"],
                            "net_from_channel_id": left["channel_id"],
                            "net_from_connector_name": left["connector_name"],
                            "net_from_cavity": left["cavity"],
                            "net_to_refdes": right["refdes"],
                            "net_to_channel_id": right["channel_id"],
                            "net_to_connector_name": right["connector_name"],
                            "net_to_cavity": right["cavity"],
                            "from_side_device_refdes": from_refdes,
                            "from_side_device_chname": from_channel_id,
                            "to_side_device_refdes": to_refdes,
                            "to_side_device_chname": to_channel_id,
                        }
                    )
                    circuit_id += 1

    # --- write circuits list ---
    with open(fileio.path("circuits list"), "w", newline="", encoding="utf-8") as f:
        writer = csv.DictWriter(f, fieldnames=COLUMNS, delimiter="\t")
        writer.writeheader()
        writer.writerows(circuits_list)



import os
import sys
import json
from pathlib import Path

from PySide6.QtWidgets import (
    QApplication,
    QWidget,
    QPushButton,
    QFileDialog,
    QMenu,
)
from PySide6.QtCore import Qt, QThread, Signal, QObject
from PySide6.QtGui import QPainter, QPen


def layout_config_path():
    """
    Save layout JSON into the root of the harnice project directory.
    """
    return Path(__file__).resolve().parents[2] / "gui_layout.json"


def run_harnice_render(cwd, lightweight=False):
    """
    Safely run harnice.cli.main() without sys.exit closing the GUI.
    """
    import harnice.cli

    old_cwd = os.getcwd()
    os.chdir(cwd)

    try:
        sys.argv = ["harnice", "-l" if lightweight else "-r"]
        try:
            harnice.cli.main()
        except SystemExit:
            pass
    finally:
        os.chdir(old_cwd)


class RenderWorker(QObject):
    finished = Signal(bool, str)  # True = success, False = error, error_message

    def __init__(self, cwd, lightweight):
        super().__init__()
        self.cwd = cwd
        self.lightweight = lightweight

    def run(self):
        try:
            run_harnice_render(self.cwd, self.lightweight)
            self.finished.emit(True, "")
        except Exception as e:
            import traceback

            error_msg = f"{type(e).__name__}: {str(e)}\n{traceback.format_exc()}"
            self.finished.emit(False, error_msg)


class GridWidget(QWidget):
    BUTTON_WIDTH = 180
    BUTTON_HEIGHT = 40
    BUTTON_WIDTH_MARGIN = 20
    BUTTON_HEIGHT_MARGIN = 20

    def __init__(self, parent=None):
        super().__init__(parent)
        self.grid_buttons = {}
        self.setStyleSheet("background-color: white;")
        self.setAutoFillBackground(True)

    @property
    def GRID_SPACING_X(self):
        return self.BUTTON_WIDTH + self.BUTTON_WIDTH_MARGIN

    @property
    def GRID_SPACING_Y(self):
        return self.BUTTON_HEIGHT + self.BUTTON_HEIGHT_MARGIN

    @property
    def OFFSET_X(self):
        return self.GRID_SPACING_X / 2

    @property
    def OFFSET_Y(self):
        return self.GRID_SPACING_Y / 2

    def paintEvent(self, event):
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)
        pen = QPen(Qt.GlobalColor.gray, 1, Qt.PenStyle.DotLine)
        painter.setPen(pen)

        width = self.width()
        height = self.height()

        x = self.OFFSET_X
        while x < width:
            painter.drawLine(x, 0, x, height)
            x += self.GRID_SPACING_X

        y = self.OFFSET_Y
        while y < height:
            painter.drawLine(0, y, width, y)
            y += self.GRID_SPACING_Y

    def grid_to_screen(self, grid_x, grid_y):
        return (
            grid_x * self.GRID_SPACING_X + self.OFFSET_X,
            grid_y * self.GRID_SPACING_Y + self.OFFSET_Y,
        )

    def screen_to_grid(self, screen_x, screen_y):
        return (
            int((screen_x - self.OFFSET_X) / self.GRID_SPACING_X),
            int((screen_y - self.OFFSET_Y) / self.GRID_SPACING_Y),
        )

    def is_grid_occupied(self, grid_x, grid_y, exclude=None):
        btn = self.grid_buttons.get((grid_x, grid_y))
        return btn is not None and btn is not exclude


class PartButton(QPushButton):
    def __init__(self, parent, label, path, grid_x, grid_y, main_window=None):
        super().__init__(label, parent)
        self.parent_grid = parent
        self.path = path
        self.grid_x = grid_x
        self.grid_y = grid_y
        self.main_window = main_window

        # ✅ Store the intended "default / unclicked" theme
        self.default_style = """
            QPushButton {
                background-color: #e6e6e6;
                border: 1px solid #666;
                border-radius: 4px;
            }
            QPushButton:hover {
                background-color: #f2f2f2;
            }
            QPushButton:pressed {
                background-color: #d0d0d0;
            }
        """
        self.setStyleSheet(self.default_style)

        self.setFixedSize(parent.BUTTON_WIDTH, parent.BUTTON_HEIGHT)
        self.dragStartPosition = None
        self.is_dragging = False
        self.show()
        self.update_position()

    def update_position(self):
        x, y = self.parent_grid.grid_to_screen(self.grid_x, self.grid_y)
        self.move(x - self.width() // 2, y - self.height() // 2)
        self.raise_()

    def mousePressEvent(self, event):
        if event.button() == Qt.MouseButton.LeftButton:
            self.dragStartPosition = event.position().toPoint()
            self.is_dragging = False
        super().mousePressEvent(event)

    def mouseMoveEvent(self, event):
        if (
            not (event.buttons() & Qt.MouseButton.LeftButton)
            or not self.dragStartPosition
        ):
            return

        if (event.position().toPoint() - self.dragStartPosition).manhattanLength() < 8:
            return

        self.is_dragging = True
        old_pos = (self.grid_x, self.grid_y)

        global_pos = self.mapToGlobal(event.position().toPoint())
        local_pos = self.parent_grid.mapFromGlobal(global_pos)
        new_x, new_y = self.parent_grid.screen_to_grid(local_pos.x(), local_pos.y())

        if (new_x, new_y) == old_pos:
            return

        if self.parent_grid.is_grid_occupied(new_x, new_y, exclude=self):
            return

        self.grid_x, self.grid_y = new_x, new_y
        self.update_position()

        self.parent_grid.grid_buttons.pop(old_pos, None)
        self.parent_grid.grid_buttons[(new_x, new_y)] = self

    def mouseReleaseEvent(self, event):
        if self.is_dragging:
            if self.main_window:
                self.main_window.save_layout()
            self.is_dragging = False
            event.accept()
            return
        super().mouseReleaseEvent(event)

    def contextMenuEvent(self, event):
        if not self.main_window:
            return

        menu = QMenu(self)
        remove = menu.addAction("Remove button")
        remove.triggered.connect(lambda: self.main_window.remove_button(self))

        newrev = menu.addAction("Create new revision")
        newrev.triggered.connect(lambda: self.main_window.new_rev(self))

        menu.exec(event.globalPos())


class HarniceGUI(QWidget):
    def __init__(self):
        super().__init__()
        self.setWindowTitle("Harnice")
        self.setWindowFlags(Qt.WindowType.WindowStaysOnTopHint | Qt.WindowType.Window)

        self.grid = GridWidget(self)

        default_width = self.grid.GRID_SPACING_X * 6
        default_height = self.grid.GRID_SPACING_Y * 2
        self.resize(default_width, default_height)
        self.grid.setGeometry(0, 0, default_width, default_height)

        self.load_button = QPushButton("Load part for render...", self.grid)
        self.load_button.setFixedSize(self.grid.BUTTON_WIDTH, self.grid.BUTTON_HEIGHT)
        self.load_button.clicked.connect(self.pick_folder)

        x, y = self.grid.grid_to_screen(0, 0)
        self.load_button.move(
            x - self.load_button.width() // 2, y - self.load_button.height() // 2
        )
        self.grid.grid_buttons[(0, 0)] = self.load_button

        self._is_initializing = True
        self.load_layout()

        # Set window size after loading layout (if any)
        self.apply_window_size()
        self._is_initializing = False

    def resizeEvent(self, event):
        self.grid.setGeometry(0, 0, self.width(), self.height())
        super().resizeEvent(event)
        # Save layout when window is manually resized
        if hasattr(self, "load_button") and not getattr(
            self, "_is_initializing", False
        ):
            self.save_layout()

    def pick_folder(self):
        folder = QFileDialog.getExistingDirectory(self, "Select revision folder")
        if not folder:
            return

        gx, gy = self.find_next_grid_position()
        label = os.path.basename(folder)

        btn = PartButton(self.grid, label, folder, gx, gy, main_window=self)
        btn.clicked.connect(lambda checked=False, p=folder: self.run_render(p))
        self.grid.grid_buttons[(gx, gy)] = btn
        self.save_layout()

    def find_next_grid_position(self):
        for y in range(200):
            for x in range(200):
                if not self.grid.is_grid_occupied(x, y):
                    return (x, y)

    def run_render(self, cwd):
        btn = next(
            (
                b
                for b in self.grid.grid_buttons.values()
                if isinstance(b, PartButton) and b.path == cwd
            ),
            None,
        )

        if btn:
            btn.setStyleSheet("background-color: #b1ffb1;")  # Green while running

        self.thread = QThread()
        self.worker = RenderWorker(cwd, False)
        self.worker.moveToThread(self.thread)

        self.thread.started.connect(self.worker.run)
        self.worker.finished.connect(
            lambda success, error_msg: self.on_render_finished(btn, success, error_msg)
        )
        self.worker.finished.connect(self.thread.quit)
        self.worker.finished.connect(self.worker.deleteLater)
        self.thread.finished.connect(self.thread.deleteLater)
        self.thread.start()

    def on_render_finished(self, btn, success, error_msg):
        if not btn:
            return

        if success:
            # ✅ Restore original intended appearance
            btn.setStyleSheet(btn.default_style)
            btn.update()
        else:
            btn.setStyleSheet("background-color: #ffb1b1;")  # Error red
            # Print error message and traceback to console in color
            if error_msg:
                # ANSI color codes
                RED = "\033[91m"
                BOLD = "\033[1m"
                RESET = "\033[0m"
                YELLOW = "\033[93m"

                print(f"\n{RED}{BOLD}{'=' * 80}{RESET}")
                print(f"{RED}{BOLD}ERROR in {YELLOW}{btn.path}{RESET}{RED}:{RESET}")
                print(f"{RED}{BOLD}{'=' * 80}{RESET}")
                print(f"{RED}{error_msg}{RESET}")
                print(f"{RED}{BOLD}{'=' * 80}{RESET}\n")

    def remove_button(self, button):
        self.grid.grid_buttons.pop((button.grid_x, button.grid_y), None)
        button.deleteLater()
        self.save_layout()

    def new_rev(self, button):
        """
        Run harnice --newrev from the button's directory.
        """
        import harnice.cli

        old_cwd = os.getcwd()
        os.chdir(button.path)

        try:
            sys.argv = ["harnice", "--newrev"]
            try:
                harnice.cli.main()
            except SystemExit:
                pass
        finally:
            os.chdir(old_cwd)

    def save_layout(self):
        data = {
            "window": {
                "width": self.width(),
                "height": self.height(),
            },
            "buttons": [
                {
                    "label": b.text(),
                    "path": b.path,
                    "grid_x": b.grid_x,
                    "grid_y": b.grid_y,
                }
                for (gx, gy), b in self.grid.grid_buttons.items()
                if isinstance(b, PartButton)
            ],
        }

        with open(layout_config_path(), "w", encoding="utf-8") as f:
            json.dump(data, f, indent=2)

    def load_layout(self):
        cfg = layout_config_path()
        if not cfg.exists():
            return

        try:
            data = json.loads(cfg.read_text(encoding="utf-8"))
        except Exception:
            return

        # Handle old format (array of buttons) vs new format (dict with buttons and window)
        if isinstance(data, list):
            items = data
            self.saved_window_size = None
        else:
            items = data.get("buttons", [])
            window_info = data.get("window")
            if window_info:
                self.saved_window_size = (
                    window_info.get("width"),
                    window_info.get("height"),
                )
            else:
                self.saved_window_size = None

        for item in items:
            btn = PartButton(
                self.grid,
                item["label"],
                item["path"],
                item["grid_x"],
                item["grid_y"],
                main_window=self,
            )
            btn.clicked.connect(
                lambda checked=False, p=item["path"]: self.run_render(p)
            )
            self.grid.grid_buttons[(item["grid_x"], item["grid_y"])] = btn

    def apply_window_size(self):
        if hasattr(self, "saved_window_size") and self.saved_window_size:
            width, height = self.saved_window_size
            self.resize(width, height)
            self.grid.setGeometry(0, 0, width, height)


def main():
    app = QApplication([])
    gui = HarniceGUI()
    gui.show()
    app.exec()


if __name__ == "__main__":
    main()



import os
import csv
import json
from harnice import fileio, state


default_desc = "CABLE, FUNCTION, ATTRIBUTES, etc."


def file_structure():
    return {
        f"{state.partnumber('pn-rev')}-attributes.json": "attributes",
        f"{state.partnumber('pn-rev')}-conductor_list.tsv": "conductor list",
    }


def generate_structure():
    pass


def render():
    # ========== Default JSON ==========
    default_attributes = {
        "jacket": {
            "properties": {
                "color": "gray",
                "material": "pvc",
                "od": "0.204in",
                "thickness": "0.028in",
            },
            "shield": {
                "properties": {"type": "foil", "coverage": "100%"},
                "drain_wire": {
                    "conductor": True,
                    "properties": {"gauge": "20AWG", "construction": "7x28"},
                    "appearance": {
                        "outline_color": "gray",
                        "slash_lines": {"direction": "RH", "color": "gray"},
                    },
                },
                "pair_1": {
                    "properties": {"twists": "12 per inch"},
                    "black": {
                        "conductor": True,
                        "properties": {
                            "insulation material": "polyethylene",
                            "od": "0.017in",
                            "gauge": "20AWG",
                            "construction": "7x28",
                            "material": "copper",
                        },
                        "appearance": {"base_color": "black"},
                    },
                    "white": {
                        "conductor": True,
                        "properties": {
                            "insulation material": "polyethylene",
                            "od": "0.017in",
                            "gauge": "20AWG",
                            "construction": "7x28",
                            "material": "copper",
                        },
                        "appearance": {"base_color": "white", "outline_color": "black"},
                    },
                },
            },
        }
    }

    attributes_path = fileio.path("attributes")

    # ========== Load or create attributes.json ==========
    if os.path.exists(attributes_path):
        try:
            with open(attributes_path, "r", encoding="utf-8") as f:
                attrs = json.load(f)
        except Exception as e:
            print(f"[WARNING] Could not load existing attributes.json: {e}")
            attrs = default_attributes
            with open(attributes_path, "w", encoding="utf-8") as f:
                json.dump(attrs, f, indent=4)
    else:
        attrs = default_attributes
        with open(attributes_path, "w", encoding="utf-8") as f:
            json.dump(attrs, f, indent=4)

    # ========== Traverse and build TSV ==========
    rows = []
    all_headers = {"appearance"}  # ensure it exists

    def recurse(obj, parent_chain=None):
        """Walk through nested dicts, recording conductor rows."""
        if parent_chain is None:
            parent_chain = []

        if isinstance(obj, dict):
            # Found conductor
            if obj.get("conductor") is True:
                props = obj.get("properties", {})
                appearance = obj.get("appearance", {})

                # Use previous parent chain to fill in context
                container = parent_chain[-2] if len(parent_chain) >= 2 else ""
                identifier = parent_chain[-1] if len(parent_chain) >= 1 else ""

                row = {"container": container, "identifier": identifier}

                for k, v in props.items():
                    row[k] = v
                    all_headers.add(k)

                # Convert appearance to JSON string (compact)
                row["appearance"] = (
                    json.dumps(
                        appearance, separators=(",", ":"), ensure_ascii=False
                    ).replace('"', "'")
                    if appearance
                    else ""
                )

                rows.append(row)

            # Continue traversing deeper
            for k, v in obj.items():
                if isinstance(v, (dict, list)):
                    recurse(v, parent_chain + [k])

        elif isinstance(obj, list):
            for item in obj:
                recurse(item, parent_chain)

    recurse(attrs)

    if not rows:
        print("[WARNING] No conductor entries found.")
        return

    # Define header order
    headers = ["container", "identifier"] + sorted(all_headers)

    # Write to TSV
    conductor_list_path = fileio.path("conductor list")
    with open(conductor_list_path, "w", newline="", encoding="utf-8") as f:
        writer = csv.DictWriter(
            f, fieldnames=headers, delimiter="\t", lineterminator="\n"
        )
        writer.writeheader()
        writer.writerows(rows)

    print(
        f"\ncable rendered successfully! wrote {len(rows)} rows to:\n{conductor_list_path}\n"
    )



import csv
from harnice import fileio

COLUMNS = [
    "instance_name", #documentation needed
    "mpn", #documentation needed
    "item_type", #documentation needed
    "lib_repo", #documentation needed
    "lib_subpath", #documentation needed
    "lib_desc", #documentation needed
    "lib_latest_rev", #documentation needed
    "lib_rev_used_here", #documentation needed
    "lib_status", #documentation needed
    "lib_releaseticket", #documentation needed
    "lib_datestarted", #documentation needed
    "lib_datemodified", #documentation needed
    "lib_datereleased", #documentation needed
    "lib_drawnby", #documentation needed
    "lib_checkedby", #documentation needed
    "project_editable_lib_modified", #documentation needed
]


def new():
    with open(fileio.path("library history"), "w", newline="", encoding="utf-8") as f:
        writer = csv.DictWriter(f, fieldnames=COLUMNS, delimiter="\t")
        writer.writeheader()
        writer.writerows([])


def append(instance_name, instance_data):
    instance_data["instance_name"] = instance_name
    for row in fileio.read_tsv("library history"):
        if row.get("instance name") == instance_name:
            raise ValueError(
                f"You're trying to import something with instance_name '{instance_name}' but it has already been imported."
            )
    with open(fileio.path("library history"), "a", newline="", encoding="utf-8") as f:
        writer = csv.DictWriter(f, fieldnames=COLUMNS, delimiter="\t")
        writer.writerow({key: instance_data.get(key, "") for key in COLUMNS})



import os
import runpy
import math
import json
import shutil
from harnice import fileio
from harnice.utils import library_utils


def run_macro(
    macro_part_number, lib_subpath, lib_repo, artifact_id, base_directory=None, **kwargs
):
    """
    Runs a macro script from the library with the given artifact ID.

    Imports a macro from the library and executes its Python script. The macro
    is pulled into a directory structure and then executed with the `artifact_id`
    and any additional keyword arguments passed as global variables.

    **Args:**
    - `macro_part_number` (str): Part number of the macro to run.
    - `lib_subpath` (str): Library subpath where the macro is located.
    - `lib_repo` (str): Library repository URL or `"local"` for local library.
    - `artifact_id` (str): Unique identifier for this macro execution (must be unique).
    - `base_directory` (str, optional): Base directory for the macro output. If `None`,
        defaults to `instance_data/macro/{artifact_id}`.
    - `**kwargs`: Additional keyword arguments to pass as global variables to the macro script.

    **Raises:**
    - `ValueError`: If `artifact_id` is `None`, `macro_part_number` is `None`, `lib_repo` is `None`,
        or if a macro with the given `artifact_id` already exists in library history.
    """
    if artifact_id is None:
        raise ValueError("artifact_id is required")
    if macro_part_number is None:
        raise ValueError("macro_part_number is required")
    if lib_repo is None:
        raise ValueError("lib_repo is required")

    for instance in fileio.read_tsv("library history"):
        if instance.get("instance_name") == artifact_id:
            raise ValueError(f"Macro with ID {artifact_id} already exists")

    if base_directory is None:
        base_directory = os.path.join("instance_data", "macro", artifact_id)

    os.makedirs(fileio.dirpath(None, base_directory), exist_ok=True)

    library_utils.pull(
        {
            "mpn": macro_part_number,
            "lib_repo": lib_repo,
            "lib_subpath": lib_subpath,
            "item_type": "macro",
            "instance_name": artifact_id,
        },
        destination_directory=fileio.dirpath(None, base_directory=base_directory),
        update_instances_list=False,
    )

    script_path = os.path.join(
        fileio.dirpath(None, base_directory=base_directory), f"{macro_part_number}.py"
    )

    # always pass the basics, but let kwargs override/extend
    init_globals = {
        "artifact_id": artifact_id,
        "artifact_path": base_directory,
        "base_directory": base_directory,
        **kwargs,  # merges/overrides
    }

    runpy.run_path(script_path, run_name="__main__", init_globals=init_globals)


def lookup_outputcsys_from_lib_used(instance, outputcsys, base_directory=None):
    """
    Looks up coordinate system transform from an instance's library attributes.

    Reads the instance's attributes JSON file to find the specified output coordinate
    system definition and returns its transform values. If the coordinate system is
    `"origin"`, returns zero transform.

    **Args:**
    - `instance` (dict): Instance dictionary containing `item_type` and `instance_name`.
    - `outputcsys` (str): Name of the output coordinate system to look up (`"origin"` returns zero transform).
    - `base_directory` (str, optional): Base directory path. If `None`, uses current working directory.

    **Returns:**
    - `tuple`: A tuple of `(x, y, rotation)` representing the coordinate system transform.
        Returns `(0, 0, 0)` if the coordinate system is `"origin"` or if the attributes
        file is not found.

    **Raises:**
    - `ValueError`: If the specified output coordinate system is not defined in the
        instance's attributes file.
    """
    if outputcsys == "origin":
        return 0, 0, 0

    attributes_path = os.path.join(
        fileio.dirpath(None, base_directory=base_directory),
        "instance_data",
        instance.get("item_type"),
        instance.get("instance_name"),
        f"{instance.get('instance_name')}-attributes.json",
    )

    try:
        with open(attributes_path, "r", encoding="utf-8") as f:
            attributes_data = json.load(f)
    except FileNotFoundError:
        return 0, 0, 0

    csys_children = attributes_data.get("csys_children", {})

    if outputcsys not in csys_children:
        raise ValueError(
            f"[ERROR] Output coordinate system '{outputcsys}' not defined in {attributes_path}"
        )

    child_csys = csys_children[outputcsys]

    # Extract values with safe numeric defaults
    x = child_csys.get("x", 0)
    y = child_csys.get("y", 0)
    angle = child_csys.get("angle", 0)
    distance = child_csys.get("distance", 0)
    rotation = child_csys.get("rotation", 0)

    # Convert angle to radians if it's stored in degrees
    angle_rad = math.radians(angle)

    # Apply translation based on distance + angle
    x = x + distance * math.cos(angle_rad)
    y = y + distance * math.sin(angle_rad)

    return x, y, rotation


def copy_pdfs_to_cwd():
    """
    Copies all PDF files from `instance_data` directory to the current working directory.

    Recursively searches the `instance_data` directory tree and copies all PDF files
    found to the current working directory. Preserves file metadata during copy.
    Prints error messages if any files cannot be copied but continues processing.
    """
    cwd = os.getcwd()

    for root, _, files in os.walk(fileio.dirpath(None, base_directory="instance_data")):
        for filename in files:
            if filename.lower().endswith(".pdf"):
                source_path = os.path.join(root, filename)
                dest_path = os.path.join(cwd, filename)

                try:
                    shutil.copy2(source_path, dest_path)  # preserves metadata
                except Exception as e:
                    print(f"[ERROR] Could not copy {source_path}: {e}")


def run_feature_for_relative(project_key, referenced_pn_rev, feature_tree_utils_name):
    """
    Runs a feature tree script from a referenced part's `features_for_relatives` directory.

    Executes a Python script located in the `features_for_relatives` directory of a
    referenced part. This is used to run feature scripts that are associated with
    parts referenced by the current project.

    **Args:**
    - `project_key` (str): Key identifying the project to look up.
    - `referenced_pn_rev` (tuple): Tuple of `(part_number, revision)` for the referenced part.
    - `feature_tree_utils_name` (str): Filename of the feature tree script to execute.
    """
    project_path = fileio.get_path_to_project(project_key)
    feature_tree_utils_path = os.path.join(
        project_path,
        f"{referenced_pn_rev[0]}-{referenced_pn_rev[1]}",
        "features_for_relatives",
        feature_tree_utils_name,
    )
    runpy.run_path(feature_tree_utils_path, run_name="__main__")



import csv
import os
from harnice import fileio

COLUMNS = [
    "merged_net", #documentation needed
    "from_device_refdes", #documentation needed
    "from_device_channel_id", #documentation needed
    "from_channel_type", #documentation needed
    "to_device_refdes", #documentation needed
    "to_device_channel_id", #documentation needed
    "to_channel_type", #documentation needed
    "multi_ch_junction_id", #documentation needed
    "disconnect_refdes_requirement", #documentation needed
    "chain_of_connectors", #documentation needed
    "chain_of_nets", #documentation needed
    "manual_map_channel_python_equiv", #documentation needed
]


def new():
    """
    Makes a new blank channel map. Overwrites existing channel map.

    Args: none

    Returns: none
    """
    channel_map = []

    for connector in fileio.read_tsv("system connector list"):
        device_refdes = connector.get("device_refdes")

        if connector.get("disconnect") == "TRUE":
            continue

        device_signals_list_path = os.path.join(
            fileio.dirpath("instance_data"),
            "device",
            device_refdes,
            f"{device_refdes}-signals_list.tsv",
        )

        for signal in fileio.read_tsv(device_signals_list_path):
            sig_channel = signal.get("channel_id")

            already = any(
                row.get("from_device_refdes") == device_refdes
                and row.get("from_device_channel_id") == sig_channel
                for row in channel_map
            )
            if already:
                continue

            if not signal.get("connector_name") == connector.get("connector"):
                continue

            channel_map_row = {
                "merged_net": connector.get("merged_net", ""),
                "from_channel_type": signal.get("channel_type", ""),
                "from_device_refdes": device_refdes,
                "from_device_channel_id": sig_channel,
            }
            channel_map.append(channel_map_row)

    # write channel map TSV
    with open(fileio.path("channel map"), "w", newline="", encoding="utf-8") as f:
        writer = csv.DictWriter(f, fieldnames=COLUMNS, delimiter="\t")
        writer.writeheader()
        writer.writerows(channel_map)

    # initialize mapped channels set TSV (empty, single column)
    with open(
        fileio.path("mapped channels set"), "w", newline="", encoding="utf-8"
    ) as f:
        pass

    return channel_map


def map(from_key, to_key=None, multi_ch_junction_key=""):
    if from_key in already_mapped_set():
        raise ValueError(f"from_key {from_key} already mapped")
    if to_key and to_key in already_mapped_set():
        raise ValueError(f"to_key {to_key} already mapped")

    channels = fileio.read_tsv("channel map")

    to_channel = None
    for channel in channels:
        if (
            channel.get("from_device_refdes") == to_key[0]
            and channel.get("from_device_channel_id") == to_key[1]
        ):
            to_channel = channel
            break

    from_channel = None
    for channel in channels:
        if (
            channel.get("from_device_refdes") == from_key[0]
            and channel.get("from_device_channel_id") == from_key[1]
        ):
            from_channel = channel
            break

    if not to_channel and multi_ch_junction_key == "":
        raise ValueError(f"to_key {to_key} not found in channel map")
    else:
        require_to = bool(to_key[0] or to_key[1])

    updated_channels, found_from, found_to = [], False, False

    for from_channel in channels:
        if (
            from_channel.get("from_device_refdes") == from_key[0]
            and from_channel.get("from_device_channel_id") == from_key[1]
        ):
            from_channel["to_device_refdes"] = to_key[0]
            from_channel["to_device_channel_id"] = to_key[1]
            from_channel["to_channel_type"] = to_channel.get("from_channel_type")
            if multi_ch_junction_key:
                from_channel["multi_ch_junction_id"] = multi_ch_junction_key
            found_from = True

            if require_to:
                from_channel["manual_map_channel_python_equiv"] = (
                    f"channel_map.map({from_key}, {to_key})"
                )
            elif multi_ch_junction_key:
                from_channel["manual_map_channel_python_equiv"] = (
                    f"channel_map.map({from_key}, multi_ch_junction_key={multi_ch_junction_key})"
                )
        elif (
            require_to
            and from_channel.get("from_device_refdes") == to_key[0]
            and from_channel.get("from_device_channel_id") == to_key[1]
        ):
            found_to = True
            continue
        updated_channels.append(from_channel)

    if not found_from:
        raise ValueError(f"from_key {from_key} not found in channel map")
    if require_to and not found_to:
        raise ValueError(f"to_key {to_key} not found in channel map")

    already_mapped_set_append(from_key)
    already_mapped_set_append(to_key)

    with open(fileio.path("channel map"), "w", newline="", encoding="utf-8") as f:
        writer = csv.DictWriter(f, fieldnames=COLUMNS, delimiter="\t")
        writer.writeheader()
        writer.writerows(updated_channels)


def already_mapped_set_append(key):
    items = already_mapped_set()
    if str(key) in items:
        raise ValueError(f"key {key} already mapped")
    items.add(str(key))
    with open(
        fileio.path("mapped channels set"), "w", newline="", encoding="utf-8"
    ) as f:
        writer = csv.writer(f, delimiter="\t")
        for item in sorted(items):
            writer.writerow([item])


def already_mapped_set():
    if not os.path.exists(fileio.path("mapped channels set")):
        return set()
    with open(fileio.path("mapped channels set"), newline="", encoding="utf-8") as f:
        reader = csv.reader(f, delimiter="\t")
        return set(row[0] for row in reader if row)


def already_mapped(key):
    if str(key) in already_mapped_set():
        return True
    else:
        return False



import os
import runpy
import sexpdata
import json
import csv
from harnice import fileio, state
from harnice.lists import signals_list, rev_history
from harnice.products import chtype

default_desc = "DEVICE, FUNCTION, ATTRIBUTES, etc."

device_feature_tree_utils_default = """
from harnice.lists import signals_list
from harnice.products import chtype

ch_type_ids = {
    "in": (1, "https://github.com/harnice/harnice"),
    "out": (4, "https://github.com/harnice/harnice"),
    "chassis": (5, "https://github.com/harnice/harnice")
}

xlr_pinout = {
    "pos": 2,
    "neg": 3,
    "chassis": 1
}

connector_mpns = {
    "XLR3F": ["in1", "in2"],
    "XLR3M": ["out1", "out2"]
}

def mpn_for_connector(connector_name):
    for mpn, conn_list in connector_mpns.items():
        if connector_name in conn_list:
            return mpn
    return None

signals_list.new()

for connector_name in ["in1", "in2", "out1", "out2"]:
    if connector_name.startswith("in"):
        channel_type = ch_type_ids["in"]
    elif connector_name.startswith("out"):
        channel_type = ch_type_ids["out"]
    else:
        continue

    channel_name = connector_name
    connector_mpn = mpn_for_connector(connector_name)

    for signal in chtype.signals(channel_type):
        signals_list.append(
            channel_id=channel_name,
            signal=signal,
            connector_name=connector_name,
            cavity=xlr_pinout.get(signal),
            channel_type=channel_type,
            connector_mpn=connector_mpn
        )

    # Add shield row
    signals_list.append(
        channel_id=f"{channel_name}-shield",
        signal="chassis",
        connector_name=connector_name,
        cavity=xlr_pinout.get("chassis"),
        channel_type=ch_type_ids["chassis"],
        connector_mpn=connector_mpn
    )

"""


def file_structure():
    return {
        f"{state.partnumber('pn-rev')}-feature_tree.py": "feature tree",
        f"{state.partnumber('pn-rev')}-signals_list.tsv": "signals list",
        f"{state.partnumber('pn-rev')}-attributes.json": "attributes",
    }


# define these here because they exist outside the rev folder you're currently working in and fileio.path cant handle that
def path(target_value):
    if target_value == "library file":
        return os.path.join(dirpath("kicad"), f"{state.partnumber('pn')}.kicad_sym")
    return fileio.path(target_value)


def dirpath(target_value):
    if target_value == "kicad":
        return os.path.join(fileio.part_directory(), "kicad")
    return fileio.dirpath(target_value)


def generate_structure():
    os.makedirs(dirpath("kicad"), exist_ok=True)


def _make_new_library_file():
    """Create a bare .kicad_sym file with only library header info."""

    symbol_lib = [
        sexpdata.Symbol("kicad_symbol_lib"),
        [sexpdata.Symbol("version"), 20241209],
        [sexpdata.Symbol("generator"), "kicad_symbol_editor"],
        [sexpdata.Symbol("generator_version"), "9.0"],
    ]

    with open(path("library file"), "w", encoding="utf-8") as f:
        sexpdata.dump(symbol_lib, f, pretty=True)


def _parse_kicad_sym_file():
    """
    Load a KiCad .kicad_sym file and return its parsed sexp data.
    """
    with open(path("library file"), "r", encoding="utf-8") as f:
        data = sexpdata.load(f)
    return data


def _symbol_exists(kicad_library_data, target_symbol_name):
    """
    Check if a symbol with a given name exists in a KiCad library.

    Args:
        kicad_library_data: Parsed sexpdata of the .kicad_sym file.
        target_symbol_name: The symbol name string to search for.

    Returns:
        True if the symbol exists, False otherwise.
    """
    for element in kicad_library_data:
        # Each element could be a list like: ["symbol", "sym_name", ...]
        if isinstance(element, list) and len(element) > 1:
            if element[0] == sexpdata.Symbol("symbol"):
                if str(element[1]) == target_symbol_name:
                    return True
    return False


def _make_property(name, value, id_counter=None, hide=False):
    # adds a property to the current rev symbol of the library
    builtins = {"Reference", "Value", "Footprint", "Datasheet", "Description"}
    prop = [
        sexpdata.Symbol("property"),
        name,
        value,  # always a string
    ]
    if name not in builtins:
        if id_counter is None:
            raise ValueError(f"Custom property {name} requires an id_counter")
        prop.append([sexpdata.Symbol("id"), id_counter])
    prop.append([sexpdata.Symbol("at"), 0, 0, 0])
    effects = [
        sexpdata.Symbol("effects"),
        [sexpdata.Symbol("font"), [sexpdata.Symbol("size"), 1.27, 1.27]],
    ]
    if hide:
        effects.append([sexpdata.Symbol("hide"), sexpdata.Symbol("yes")])
    prop.append(effects)
    return prop


def _add_blank_symbol(sym_name, value="", footprint="", datasheet="", description=""):
    """Append a blank symbol into the .kicad_sym at fileio.path('library file')."""

    lib_path = path("library file")

    # Load the existing s-expression
    with open(lib_path, "r", encoding="utf-8") as f:
        data = sexpdata.load(f)

    # Build symbol s-expression
    if rev_history.info(field="library_repo") in ["", None]:
        lib_repo_to_write = "local"
    else:
        lib_repo_to_write = _get_attribute("library_repo")
    symbol = [
        sexpdata.Symbol("symbol"),
        sym_name,
        [sexpdata.Symbol("exclude_from_sim"), sexpdata.Symbol("no")],
        [sexpdata.Symbol("in_bom"), sexpdata.Symbol("yes")],
        [sexpdata.Symbol("on_board"), sexpdata.Symbol("yes")],
        _make_property("Reference", _get_attribute("default_refdes")),
        _make_property("Value", value),
        _make_property("Footprint", footprint, hide=True),
        _make_property("Datasheet", datasheet, hide=True),
        _make_property("Description", _get_attribute("desc"), hide=True),
        _make_property("MFG", _get_attribute("manufacturer"), hide=False, id_counter=0),
        _make_property(
            "MPN", _get_attribute("manufacturer_part_number"), hide=False, id_counter=1
        ),
        _make_property("lib_repo", lib_repo_to_write, hide=True, id_counter=2),
        _make_property(
            "lib_subpath", _get_attribute("library_subpath"), hide=True, id_counter=3
        ),
        _make_property("rev", state.partnumber("rev"), hide=True, id_counter=4),
        [sexpdata.Symbol("embedded_fonts"), sexpdata.Symbol("no")],
    ]

    # Append to the library data
    data.append(symbol)

    # Write back out
    with open(lib_path, "w", encoding="utf-8") as f:
        sexpdata.dump(data, f, pretty=True)


def _overwrite_or_create_property_in_symbol(prop_name, value, hide=False):
    """
    Overwrite or create a property inside the target symbol block
    in the KiCad .kicad_sym library file.

    - File is always path("library file")
    - Symbol to modify is always named state.partnumber("pn-rev")

    Args:
        prop_name (str): Name of the property
        value (str): Value to set (will always be forced to string)
        hide (bool): Whether to hide the property
    """

    target_symbol_name = state.partnumber("pn-rev")

    # Ensure value is a string (KiCad requirement)
    if value is None:
        value = ""
    else:
        value = str(value)

    # Load the library file
    with open(path("library file"), "r", encoding="utf-8") as f:
        data = sexpdata.load(f)

    def next_id(symbol):
        """Find the next available id number among custom properties."""
        max_id = -1
        for elem in symbol:
            if (
                isinstance(elem, list)
                and len(elem) >= 4
                and isinstance(elem[0], sexpdata.Symbol)
                and elem[0].value() == "property"
            ):
                for sub in elem:
                    if isinstance(sub, list) and len(sub) == 2:
                        if (
                            isinstance(sub[0], sexpdata.Symbol)
                            and sub[0].value() == "id"
                            and isinstance(sub[1], int)
                        ):
                            max_id = max(max_id, sub[1])
        return max_id + 1

    def overwrite_or_create(symbol):
        # Try to overwrite existing property
        for elem in symbol:
            if (
                isinstance(elem, list)
                and len(elem) >= 3
                and isinstance(elem[0], sexpdata.Symbol)
                and elem[0].value() == "property"
                and elem[1] == prop_name
            ):
                elem[2] = value  # force overwrite as string
                return symbol

        # If missing, create new one with next id
        new_id = next_id(symbol)
        new_prop = _make_property(prop_name, value, id_counter=new_id, hide=hide)
        symbol.append(new_prop)
        return symbol

    # Traverse to the right (symbol ...) block
    for i, elem in enumerate(data):
        if (
            isinstance(elem, list)
            and isinstance(elem[0], sexpdata.Symbol)
            and elem[0].value() == "symbol"
            and elem[1] == target_symbol_name
        ):
            data[i] = overwrite_or_create(elem)

    # Save file back
    with open(path("library file"), "w", encoding="utf-8") as f:
        sexpdata.dump(data, f)


def _extract_pins_from_symbol(kicad_lib, symbol_name):
    """
    Extract all pin info for the given symbol (and its subsymbols).
    Returns a list of dicts like {"name": ..., "number": ..., "type": ..., "shape": ...}.
    """

    def sym_to_str(obj):
        """Convert sexpdata.Symbol to string, pass through everything else."""
        if isinstance(obj, sexpdata.Symbol):
            return obj.value()
        return obj

    pins = []

    def recurse(node, inside_target=False):
        if not isinstance(node, list) or not node:
            return

        tag = sym_to_str(node[0])

        if tag == "symbol":
            sym_name = sym_to_str(node[1])
            new_inside = inside_target or (sym_name == symbol_name)
            for sub in node[2:]:
                recurse(sub, inside_target=new_inside)

        elif tag == "pin" and inside_target:
            pin_type = sym_to_str(node[1]) if len(node) > 1 else None
            pin_shape = sym_to_str(node[2]) if len(node) > 2 else None
            name_val = None
            number_val = None

            for entry in node[3:]:
                if isinstance(entry, list) and entry:
                    etag = sym_to_str(entry[0])
                    if etag == "name":
                        name_val = sym_to_str(entry[1])
                    elif etag == "number":
                        number_val = sym_to_str(entry[1])

            pin_info = {
                "name": name_val,
                "number": number_val,
                "type": pin_type,
                "shape": pin_shape,
            }
            pins.append(pin_info)

        else:
            for sub in node[1:]:
                recurse(sub, inside_target=inside_target)

    recurse(kicad_lib, inside_target=False)
    return pins


def _validate_pins(pins, unique_connectors_in_signals_list):
    """Validate pins for uniqueness, type conformity, and check required pins.

    Returns:
        tuple:
            missing (set): Any missing pin names from unique_connectors_in_signals_list.
            used_pin_numbers (set): Numbers already assigned to pins.
    Raises:
        ValueError: On duplicate names/numbers or invalid types.
    """
    seen_names = set()
    seen_numbers = set()

    for pin in pins:
        name = pin.get("name")
        number = pin.get("number")
        ptype = pin.get("type")

        # Duplicate name
        if name in seen_names:
            raise ValueError(f"Duplicate pin name found: {name}")
        seen_names.add(name)

        # Duplicate number
        if number in seen_numbers:
            raise ValueError(f"Duplicate pin number found: {number}")
        seen_numbers.add(number)

        # Type check
        if ptype != "unspecified":
            raise ValueError(
                f"Pin {name} ({number}) has invalid type: {ptype}. Harnice requires all pins to have type 'unspecified'."
            )

    # Set comparison for 1:1 match
    required = set(unique_connectors_in_signals_list)
    pin_names = seen_names

    missing = required - pin_names
    extra = pin_names - required
    if extra:
        raise ValueError(
            f"The following pin(s) exist in KiCad symbol but not Signals List: {', '.join(sorted(extra))}"
        )

    return missing, seen_numbers


def _append_missing_pin(pin_name, pin_number, spacing=3.81):
    """
    Append a pin to the KiCad symbol whose name matches state.partnumber('pn-rev').
    Immediately writes the updated symbol back to path("library file").
    """
    file_path = path("library file")
    pin_number = str(pin_number)
    target_name = state.partnumber("pn-rev")

    import sexpdata

    # --- Load file ---
    with open(file_path, "r", encoding="utf-8") as f:
        symbol_data = sexpdata.load(f)

    # --- Find the symbol with matching name ---
    target_symbol = None
    for item in symbol_data:
        if (
            isinstance(item, list)
            and len(item) >= 2
            and isinstance(item[0], sexpdata.Symbol)
            and item[0].value() == "symbol"
        ):
            # (symbol "Name" ...)
            name_token = item[1]
            if isinstance(name_token, str) and name_token.strip() == target_name:
                target_symbol = item
                break

    if target_symbol is None:
        raise ValueError(f"No symbol named '{target_name}' found in {file_path}")

    # --- Skip if duplicate already present ---
    for elem in target_symbol[2:]:
        if (
            isinstance(elem, list)
            and isinstance(elem[0], sexpdata.Symbol)
            and elem[0].value() == "pin"
        ):
            name_entry = next(
                (
                    x
                    for x in elem
                    if isinstance(x, list)
                    and len(x) > 1
                    and isinstance(x[0], sexpdata.Symbol)
                    and x[0].value() == "name"
                ),
                None,
            )
            num_entry = next(
                (
                    x
                    for x in elem
                    if isinstance(x, list)
                    and len(x) > 1
                    and isinstance(x[0], sexpdata.Symbol)
                    and x[0].value() == "number"
                ),
                None,
            )
            if (
                name_entry
                and name_entry[1] == pin_name
                and num_entry
                and num_entry[1] == pin_number
            ):
                return symbol_data  # already present

    # --- Find max Y among existing pins ---
    max_y = -spacing
    for elem in target_symbol[2:]:
        if (
            isinstance(elem, list)
            and isinstance(elem[0], sexpdata.Symbol)
            and elem[0].value() == "pin"
        ):
            at_entry = next(
                (
                    x
                    for x in elem
                    if isinstance(x, list)
                    and len(x) >= 3
                    and isinstance(x[0], sexpdata.Symbol)
                    and x[0].value() == "at"
                ),
                None,
            )
            if at_entry:
                y_val = float(at_entry[2])
                max_y = max(max_y, y_val)

    new_y = max_y + spacing

    # --- Build new pin ---
    new_pin = [
        sexpdata.Symbol("pin"),
        sexpdata.Symbol("unspecified"),
        sexpdata.Symbol("line"),
        [sexpdata.Symbol("at"), 0, new_y, 0],
        [sexpdata.Symbol("length"), 2.54],
        [
            sexpdata.Symbol("name"),
            pin_name,
            [
                sexpdata.Symbol("effects"),
                [sexpdata.Symbol("font"), [sexpdata.Symbol("size"), 1.27, 1.27]],
            ],
        ],
        [
            sexpdata.Symbol("number"),
            pin_number,
            [
                sexpdata.Symbol("effects"),
                [sexpdata.Symbol("font"), [sexpdata.Symbol("size"), 1.27, 1.27]],
            ],
        ],
    ]

    target_symbol.append(new_pin)

    # --- Write back ---
    with open(file_path, "w", encoding="utf-8") as f:
        sexpdata.dump(symbol_data, f)

    print(
        f"Appended pin {pin_name} ({pin_number}) to symbol '{target_name}' in {os.path.basename(file_path)}"
    )
    return symbol_data


def _remove_details_from_signals_list():
    """Remove the specified channel-related columns from the signals list."""
    old_list = fileio.read_tsv("signals list")

    COLUMNS_TO_DROP = {"channel_id", "signal", "cavity"}

    new_list = []
    for row in old_list:
        filtered = {k: v for k, v in row.items() if k not in COLUMNS_TO_DROP}
        new_list.append(filtered)

    # Rewrite the TSV
    path = fileio.path("signals list")
    with open(path, "w", newline="", encoding="utf-8") as f:
        writer = csv.DictWriter(f, fieldnames=signals_list.COLUMNS, delimiter="\t")
        writer.writeheader()
        writer.writerows(new_list)


def _next_free_number(seen_numbers, start=1):
    """Find the next unused pin number as a string."""
    n = start
    while True:
        if str(n) not in seen_numbers:
            return str(n)
        n += 1


def _validate_kicad_library():
    """
    Validate that the KiCad .kicad_sym library has:
    0. The .kicad_sym file exists (create if missing).
    1. A symbol matching the current part number.
    2. Pins that match the connectors in the signals list.
    """

    if not os.path.exists(path("library file")):
        _make_new_library_file()
        print("New Kicad symbol file created.")

    kicad_library_data = _parse_kicad_sym_file()

    if not _symbol_exists(kicad_library_data, state.partnumber("pn-rev")):
        _add_blank_symbol(
            sym_name=state.partnumber("pn-rev"),
        )

    # Step 1. Collect unique connectors from the signals list
    unique_connectors_in_signals_list = set()
    for signal in fileio.read_tsv("signals list"):
        connector_name = signal.get("connector_name")
        if connector_name:
            unique_connectors_in_signals_list.add(connector_name)

    # Step 2. Validate pins
    kicad_lib = _parse_kicad_sym_file()
    pins = _extract_pins_from_symbol(kicad_lib, state.partnumber("pn-rev"))
    missing, seen_numbers = _validate_pins(pins, unique_connectors_in_signals_list)

    kicad_library_data = _parse_kicad_sym_file()

    # Step 3. Append missing pins
    for pin_name in missing:
        # find the next available number
        pin_number = _next_free_number(seen_numbers)
        # append it
        _append_missing_pin(pin_name, pin_number)
        # mark number as used
        seen_numbers.add(pin_number)

    # Step 4. Overwrite symbol properties
    _overwrite_or_create_property_in_symbol(
        "Reference", _get_attribute("default_refdes"), hide=False
    )
    _overwrite_or_create_property_in_symbol(
        "Description", _get_attribute("desc"), hide=False
    )
    _overwrite_or_create_property_in_symbol("MFG", _get_attribute("mfg"), hide=True)
    _overwrite_or_create_property_in_symbol("MPN", _get_attribute("pn"), hide=False)

    if rev_history.info(field="library_repo") in ["", None]:
        _overwrite_or_create_property_in_symbol("lib_repo", "local", hide=True)
    else:
        _overwrite_or_create_property_in_symbol(
            "lib_repo", _get_attribute("library_repo"), hide=True
        )
    _overwrite_or_create_property_in_symbol(
        "lib_subpath", _get_attribute("library_subpath"), hide=True
    )
    _overwrite_or_create_property_in_symbol("rev", state.partnumber("rev"), hide=True)


def _validate_attributes_json():
    """Ensure an attributes JSON file exists with default values if missing."""

    default_attributes = {"default_refdes": "DEVICE"}

    attributes_path = fileio.path("attributes")

    # If attributes file does not exist, create it with defaults
    if not os.path.exists(attributes_path):
        with open(attributes_path, "w", encoding="utf-8") as f:
            json.dump(default_attributes, f, indent=4)

    # If it exists, load it and verify required keys
    else:
        with open(attributes_path, "r", encoding="utf-8") as f:
            try:
                attributes = json.load(f)
            except json.JSONDecodeError:
                raise ValueError(f"Invalid JSON in attributes file: {attributes_path}")

        updated = False
        for key, value in default_attributes.items():
            if key not in attributes:
                attributes[key] = value
                updated = True

        if updated:
            with open(attributes_path, "w", encoding="utf-8") as f:
                json.dump(attributes, f, indent=4)
            print(f"Updated attributes file with missing defaults at {attributes_path}")


def _get_attribute(attribute_key):
    # find an attribute from either revision history tsv or attributes json
    if attribute_key in rev_history.COLUMNS:
        return rev_history.info(field=attribute_key)

    else:
        with open(fileio.path("attributes"), "r", encoding="utf-8") as f:
            return json.load(f).get(attribute_key)


def configurations(sig_list):
    """
    Returns a dict of each configuration variable and each of its allowed options.
    {number} represents any number and can be used in a string like "{number}V".
    You can also say "0<={number}<10V" to describe bounds.

    Args:
    signals_list (dictionary form)

    Returns:
        {
            "config_col_1": {"opt1", "opt2", ""},
            "config_col_2": {"5V", "12V", ""},
            ...
        }
    """

    # collect headers
    headers = set()
    for item in sig_list:
        headers.update(item.keys())
        break  # only need first row for headers

    # find configuration columns
    configuration_cols = (
        headers - set(signals_list.DEVICE_COLUMNS) - {"config_variable"}
    )

    # initialize root dict
    configuration_vars = {col: set() for col in configuration_cols}

    # populate unique values (INCLUDING blanks)
    for row in sig_list:
        for col in configuration_cols:
            val = row.get(col)

            # normalize everything to string
            if val is None:
                val = ""
            else:
                val = str(val).strip()

            configuration_vars[col].add(val)

    return configuration_vars


def _validate_signals_list():
    print("--------------------------------")
    print("Validating signals list...")
    if not os.path.exists(fileio.path("signals list")):
        raise FileNotFoundError("Signals list was not generated.")

    with open(fileio.path("signals list"), "r", encoding="utf-8") as f:
        reader = csv.DictReader(f, delimiter="\t")
        headers = reader.fieldnames
        signals_list = list(reader)

    if not headers:
        raise ValueError("Signals list has no header row.")

    config_vars = configurations(signals_list)

    print(json.dumps(config_vars, indent=4))
    # NEXT UP: WAIT UNTIL YOU HAVE A GOOD USE CASE OF CONFIGURED DEVICES.
    # CONFIRM THAT THIS PRINTS A DICTIONARY OF ALL THE VALID CONFIGURATION VARIABLES AND THEIR DEFINED STATES
    # THEN MAKE A LIST OF EVERY SINGLE FACTORIAL COMBINATION OF THE CONFIGURATION VARIABLES
    # THEN ITERATE THROUGH THAT LIST AND VALIDATE EACH CONFIGURATION

    counter = 2
    for signal in signals_list:
        print("Looking at csv row:", counter)
        channel_type = chtype.parse(signal.get("channel_type"))
        expected_signals = chtype.signals(channel_type)
        found_signals = set()
        connector_names = set()

        # make sure all the fields are there
        if signal.get("channel_id") in ["", None]:
            raise ValueError("channel_id is blank")
        if signal.get("signal") in ["", None]:
            raise ValueError("signal is blank")
        if signal.get("connector_name") in ["", None]:
            raise ValueError("connector_name is blank")
        if signal.get("cavity") in ["", None]:
            raise ValueError("cavity is blank")
        if signal.get("connector_mpn") in ["", None]:
            raise ValueError("connector_mpn is blank")
        if signal.get("channel_type") in ["", None]:
            raise ValueError("channel_type is blank")

        # make sure signal is a valid signal of its channel type
        if signal.get("signal") not in chtype.signals(channel_type):
            raise ValueError(
                f"Signal {signal.get('signal')} is not a valid signal of its channel type"
            )

        # make sure all the signals of each channel type are present
        for expected_signal in expected_signals:
            for signal2 in signals_list:
                if (
                    signal2.get("channel_id") == signal.get("channel_id")
                    and signal2.get("signal") == expected_signal
                ):
                    found_signals.add(expected_signal)
                    connector_names.add(signal2.get("connector_name"))

        missing_signals = set(expected_signals) - found_signals
        if missing_signals:
            raise ValueError(
                f"Channel {signal.get('channel_id')} is missing signals: {', '.join(missing_signals)}"
            )

        # make sure no channels are spread across multiple connectors
        if len(connector_names) > 1:
            raise ValueError(
                f"Channel {signal.get('channel_id')} has signals spread across multiple connectors: "
                f"{', '.join(connector_names)}"
            )

        counter += 1

    # make sure no duplicate cavities are present
    seen_cavities = set()
    for signal in signals_list:
        cavity_key = f"{signal.get('connector_name')}-{signal.get('cavity')}"
        if cavity_key in seen_cavities:
            raise ValueError(
                f"Duplicate cavity '{signal.get('cavity')}' found on connector '{signal.get('connector_name')}'"
            )
        seen_cavities.add(cavity_key)

    print(f"Signals list of {state.partnumber('pn')} is valid.\n")


def render(lightweight=False):
    signals_list.set_list_type("device")
    _validate_attributes_json()

    # make a new signals list
    if not os.path.exists(fileio.path("signals list")):
        if lightweight:
            signals_list.new()
            signals_list.write_signal(
                connector_name="J1",
                channel_type=0,
                signal="placeholder",
                cavity=1,
                connector_mpn="DB9_F",
            )
        else:
            with open(fileio.path("feature tree"), "w", encoding="utf-8") as f:
                f.write(device_feature_tree_utils_default)

    if os.path.exists(fileio.path("feature tree")):
        runpy.run_path(fileio.path("feature tree"))
        print("Successfully rebuilt signals list per feature tree.")

    if not lightweight:
        _validate_signals_list()

    if lightweight:
        # don't want to map things that have not been mapped completely yet
        _remove_details_from_signals_list()

    path_nickname = ""
    subpath_nickname = ""
    if rev_history.info(field="library_repo"):
        path_nickname = f"{os.path.basename(rev_history.info(field='library_repo'))}/"
    if rev_history.info(field="library_subpath"):
        subpath_nickname = f"{rev_history.info(field='library_subpath')}"

    if path_nickname == "":
        print(
            "Add this to 'PROJECT SPECIFIC LIBRARIES' not 'global libraries' in Kicad because it doesn't look like you're working in a Harnice library path"
        )

    print(
        "\n"
        f"Nickname:       {path_nickname}{subpath_nickname}{state.partnumber('pn')}\n"
        f"Library path:   {path('library file')}\n"
    )

    _validate_kicad_library()



import csv
import os
from harnice import fileio

COLUMNS = [
    "segment_id", #documentation needed
    "node_at_end_a", #documentation needed
    "node_at_end_b", #documentation needed
    "length", #documentation needed
    "angle", #documentation needed
    "diameter", #documentation needed
]


def new():
    with open(
        fileio.path("formboard graph definition"),
        "w",
        newline="",
        encoding="utf-8",
    ) as f:
        writer = csv.DictWriter(
            f, fieldnames=COLUMNS, delimiter="\t", lineterminator="\n"
        )
        writer.writeheader()


def append(segment_id, segment_data):
    if not segment_id:
        raise ValueError(
            "Argument 'segment_id' is blank and required to identify a unique segment"
        )

    segment_data["segment_id"] = segment_id

    # Prevent duplicates
    if any(
        row.get("segment_id") == segment_id
        for row in fileio.read_tsv("formboard graph definition")
    ):
        return True

    # Ensure the file exists
    path = fileio.path("formboard graph definition")
    if not os.path.exists(path):
        new()

    # Append safely
    with open(path, "a+", newline="", encoding="utf-8") as f:
        # ---- Ensure file ends with a newline before writing ----
        f.seek(0, os.SEEK_END)
        if f.tell() > 0:  # file is non-empty
            f.seek(f.tell() - 1)
            if f.read(1) != "\n":
                f.write("\n")
        # --------------------------------------------------------

        writer = csv.DictWriter(
            f, fieldnames=COLUMNS, delimiter="\t", lineterminator="\n"
        )
        writer.writerow({key: segment_data.get(key, "") for key in COLUMNS})

    return False



import ast
import webcolors

"""
Appearance guide:
    color lookup: https://www.w3.org/TR/SVG11/types.html#ColorKeywords
    add the following dictionary-formatted string to a field in the instances list


format: design your macros to be able to parse this, define your cables to conform to this
    {"base_color":"","parallelstripe":["",""],"perpstripe":[], "twisted":None}

    
"""


def parse(val):
    """
    Parse appearance dictionary, converting color names and shorthand hex to full hex.
    No validation or safeguards.
    """
    if not val:
        return None

    data = val if isinstance(val, dict) else ast.literal_eval(str(val))
    result = {}

    for key, value in data.items():
        # lists → normalize each element
        if isinstance(value, list):
            parsed = []
            for c in value:
                c = c.strip().lower()
                if c.startswith("#") and len(c) == 4:
                    c = "#" + "".join(ch * 2 for ch in c[1:])
                if not c.startswith("#"):
                    c = webcolors.name_to_hex(c)
                parsed.append(c)
            result[key] = parsed
        # single string → normalize directly
        elif isinstance(value, str):
            c = value.strip().lower()
            if c.startswith("#") and len(c) == 4:
                c = "#" + "".join(ch * 2 for ch in c[1:])
            if not c.startswith("#"):
                c = webcolors.name_to_hex(c)
            result[key] = c
        else:
            result[key] = value

    return result



import os
import csv
from harnice import fileio

COLUMNS = [
    "net", #documentation needed
    "harness_pn", #documentation needed
]


def new():
    """
    Synchronize the system harness manifest with the system connector list:
      - Remove nets that no longer exist in the connector list
      - Add nets that appear in the connector list but not yet in the manifest
      - Preserve all other column data for nets that still exist
    """
    # Load connector list and extract unique nets
    connector_nets = {
        row.get("net", "").strip()
        for row in fileio.read_tsv("system connector list")
        if row.get("net")
    }

    manifest_path = fileio.path("harness manifest")

    # Load existing manifest if present
    existing_manifest = []
    manifest_nets = set()
    try:
        existing_manifest = fileio.read_tsv("harness manifest")
        manifest_nets = {
            row.get("net", "").strip() for row in existing_manifest if row.get("net")
        }
    except FileNotFoundError:
        existing_manifest = []
        manifest_nets = set()

    # Determine differences
    nets_to_add = connector_nets - manifest_nets
    nets_to_keep = manifest_nets & connector_nets

    # Preserve existing info for kept nets
    updated_manifest = [
        row for row in existing_manifest if row.get("net") in nets_to_keep
    ]

    # Add new rows for new nets
    for net in sorted(nets_to_add):
        updated_manifest.append({"net": net})

    # Sort by net name for consistency
    updated_manifest = sorted(updated_manifest, key=lambda r: r.get("net", ""))

    # Write updated manifest
    with open(manifest_path, "w", newline="", encoding="utf-8") as f:
        writer = csv.DictWriter(f, fieldnames=COLUMNS, delimiter="\t")
        writer.writeheader()
        for row in updated_manifest:
            full_row = {col: row.get(col, "") for col in COLUMNS}
            writer.writerow(full_row)


def update_upstream(path_to_system_rev, system_pn_rev, manifest_nets, harness_pn):
    manifest_path = os.path.join(
        path_to_system_rev,
        "lists",
        f"{system_pn_rev[0]}-{system_pn_rev[1]}-harness_manifest.tsv",
    )

    # Read existing manifest
    with open(manifest_path, newline="", encoding="utf-8") as f:
        reader = csv.DictReader(f, delimiter="\t")
        manifest = list(reader)
        fieldnames = reader.fieldnames

    # --- Pass 1: update matching nets ---
    for net in manifest_nets:
        for row in manifest:
            if row.get("net") == net:
                row["harness_pn"] = harness_pn
                break

    # --- Pass 2: remove outdated links ---
    for row in manifest:
        if row.get("harness_pn") == harness_pn and row.get("net") not in manifest_nets:
            row["harness_pn"] = ""

    # Write back updated manifest
    with open(manifest_path, "w", newline="", encoding="utf-8") as f:
        writer = csv.DictWriter(f, fieldnames=fieldnames, delimiter="\t")
        writer.writeheader()
        writer.writerows(manifest)



import os
import random
import math
import ast
from collections import defaultdict, deque
from PIL import Image, ImageDraw, ImageFont
from harnice import fileio
from harnice.lists import instances_list, formboard_graph


def validate_nodes():
    """
    Validates and initializes the formboard graph structure.

    This comprehensive function performs multiple tasks:

    1. Ensures the formboard graph definition TSV exists (creates if missing)
    2. Synchronizes nodes from the instances list with the formboard graph
    3. Creates segments if the graph is empty (wheel-spoke for >2 nodes, single segment for 2 nodes)
    4. Adds missing nodes to the graph definition
    5. Removes obsolete nodes and segments
    6. Validates graph structure (no loops, no dangling nodes, single connected component)
    7. Generates node coordinates by propagating from origin
    8. Calculates average node angles based on connected segments
    9. Generates a PNG visualization of the formboard graph

    **Raises:**

    - `ValueError`: If fewer than two nodes are defined.
    - `Exception`: If loops are detected, if nodes have no segments, or if the graph
        has disconnected components.

    """
    # Ensure TSV exists
    if not os.path.exists(fileio.path("formboard graph definition")):
        formboard_graph.new()

    # Collect node names from the instance list
    nodes_from_instances_list = {
        instance.get("instance_name")
        for instance in fileio.read_tsv("instances list")
        if instance.get("item_type") == "node"
    }

    # --- Case 1: No segments exist in formboard definition yet, build from scratch ---
    if not fileio.read_tsv("formboard graph definition"):
        # If there are more than two nodes, make a randomized wheel-spoke graph
        if len(nodes_from_instances_list) > 2:
            origin_node = "node1"
            node_counter = 0
            for instance in fileio.read_tsv("instances list"):
                if instance.get("item_type") == "node":
                    segment_id = instance.get("instance_name") + "_leg"
                    formboard_graph.append(
                        segment_id,
                        {
                            "node_at_end_a": (
                                instance.get("instance_name")
                                if node_counter == 0
                                else origin_node
                            ),
                            "node_at_end_b": (
                                origin_node
                                if node_counter == 0
                                else instance.get("instance_name")
                            ),
                            "length": str(random.randint(6, 18)),
                            "angle": str(
                                0 if node_counter == 0 else random.randint(0, 359)
                            ),
                            "diameter": 0.1,
                        },
                    )
                    node_counter += 1
        # If there are exactly two nodes, make a single segment between them
        elif len(nodes_from_instances_list) == 2:
            segment_id = "segment"
            segment_ends = []
            for instance in fileio.read_tsv("instances list"):
                if instance.get("item_type") == "node":
                    segment_ends.append(instance.get("instance_name"))

            formboard_graph.append(
                segment_id,
                {
                    "segment_id": segment_id,
                    "node_at_end_a": segment_ends[0],
                    "node_at_end_b": segment_ends[1],
                    "length": str(random.randint(6, 18)),
                    "angle": str(0),
                    "diameter": 0.1,
                },
            )
        # If there are fewer than two nodes, raise an error
        else:
            raise ValueError("Fewer than two nodes defined, cannot build segments.")

    # --- Case 2: Some nodes from instances list exist in the formboard definition but not all of them ---
    # Extract nodes already involved in segments in formboard definition
    nodes_from_formboard_definition = set()
    for row in fileio.read_tsv("formboard graph definition"):
        nodes_from_formboard_definition.add(row.get("node_at_end_a", ""))
        nodes_from_formboard_definition.add(row.get("node_at_end_b", ""))
    nodes_from_formboard_definition.discard("")

    nodes_from_instances_list_not_in_formboard_def = (
        nodes_from_instances_list - nodes_from_formboard_definition
    )

    if nodes_from_instances_list_not_in_formboard_def:
        for missing_node in nodes_from_instances_list_not_in_formboard_def:
            segment_id = f"{missing_node}_leg"

            node_to_attach_new_leg_to = ""
            for instance in fileio.read_tsv("instances list"):
                if (
                    instance.get("item_type") == "node"
                    and instance.get("instance_name") != missing_node
                ):
                    node_to_attach_new_leg_to = instance.get("instance_name")
                    break

            if not node_to_attach_new_leg_to:
                raise ValueError(
                    f"No existing node found to connect {missing_node} to."
                )

            formboard_graph.append(
                segment_id,
                {
                    "segment_id": segment_id,
                    "node_at_end_a": missing_node,
                    "node_at_end_b": node_to_attach_new_leg_to,
                    "length": str(random.randint(6, 18)),
                    "angle": str(random.randint(0, 359)),
                    "diameter": 0.1,
                },
            )

    # --- Remove any segments that connect to nodes that are both...
    #        only referenced once in the formboard definition and
    #        not in the instances list
    # remove their nodes as well

    segments = fileio.read_tsv("formboard graph definition")
    node_occurrences = defaultdict(int)

    for seg in segments:
        node_occurrences[seg.get("node_at_end_a")] += 1
        node_occurrences[seg.get("node_at_end_b")] += 1

    obsolete_nodes = [
        node
        for node, count in node_occurrences.items()
        if count == 1 and node not in nodes_from_instances_list
    ]

    if obsolete_nodes:
        cleaned_segments = [
            seg
            for seg in segments
            if seg.get("node_at_end_a") not in obsolete_nodes
            and seg.get("node_at_end_b") not in obsolete_nodes
        ]

        # reset the file to header only
        formboard_graph.new()

        # re-append each segment cleanly using your existing append logic
        for seg in cleaned_segments:
            formboard_graph.append(seg["segment_id"], seg)

    # --- Ensure each valid segment from formboard definition is represented in instances list
    for segment in fileio.read_tsv("formboard graph definition"):
        instances_list.new_instance(
            segment.get("segment_id"),
            {
                "item_type": "segment",
                "location_type": "segment",
                "segment_group": segment.get("segment_id"),
                "length": segment.get("length"),
                "diameter": segment.get("diameter"),
                "parent_csys_instance_name": segment.get("node_at_end_a"),
                "parent_csys_outputcsys_name": "origin",
                "node_at_end_a": segment.get("node_at_end_a"),
                "node_at_end_b": segment.get("node_at_end_b"),
                "absolute_rotation": segment.get("angle"),
            },
        )

    # === Ensure each node is represented in instances list ===
    for node in nodes_from_formboard_definition:
        instances_list.new_instance(
            node,
            {
                "item_type": "node",
                "location_type": "node",
                "parent_csys_instance_name": "origin",
                "parent_csys_outputcsys_name": "origin",
            },
            ignore_duplicates=True,
        )

    # === Detect loops ===
    adjacency = defaultdict(list)
    for instance in fileio.read_tsv("instances list"):
        if instance.get("item_type") == "segment":
            node_a = instance.get("node_at_end_a")
            node_b = instance.get("node_at_end_b")
            if node_a and node_b:
                adjacency[node_a].append(node_b)
                adjacency[node_b].append(node_a)

    visited = set()

    def dfs(node, parent):
        visited.add(node)
        for neighbor in adjacency[node]:
            if neighbor not in visited:
                if dfs(neighbor, node):
                    return True
            elif neighbor != parent:
                return True
        return False

    for node in adjacency:
        if node not in visited:
            if dfs(node, None):
                raise Exception(
                    "Loop detected in formboard graph. Would be cool, but Harnice doesn't support that yet."
                )

    # === Find nodes that are not connected to any segments ===
    all_nodes_in_segments = set(adjacency.keys())
    nodes_without_segments = nodes_from_instances_list - all_nodes_in_segments
    if nodes_without_segments:
        raise Exception(
            f"Dangling nodes with no connections found: {', '.join(sorted(nodes_without_segments))}"
        )

    # === Ensure each segment is part of the same graph (i.e. no disconnected components) ===
    def bfs(start):
        q = deque([start])
        seen = {start}
        while q:
            n = q.popleft()
            for nbr in adjacency.get(n, []):
                if nbr not in seen:
                    seen.add(nbr)
                    q.append(nbr)
        return seen

    all_nodes = set(adjacency.keys())
    seen_global = set()
    components = []

    for n in all_nodes:
        if n not in seen_global:
            component = bfs(n)
            seen_global |= component
            components.append(component)

    if len(components) > 1:
        formatted_connector_groups = "\n".join(
            f"  - [{', '.join(sorted(c))}]" for c in components
        )
        raise Exception(
            f"Disconnected formboard graph found ({len(components)} connector_groups):\n{formatted_connector_groups}"
        )

    # GENERATE NODE COORDINATES

    # === Step 1: Reload from instances_list ===
    instances = fileio.read_tsv("instances list")

    segments = [inst for inst in instances if inst.get("item_type") == "segment"]
    nodes = [inst for inst in instances if inst.get("item_type") == "node"]

    # === Step 2: Determine origin node ===
    origin_node = ""
    for seg in segments:
        origin_node = seg.get("node_at_end_a")
        if origin_node:
            break

    print(f"-origin node: '{origin_node}'")

    # === Step 3: Build graph from segments ===
    graph = {}
    for seg in segments:
        a = seg.get("node_at_end_a")
        b = seg.get("node_at_end_b")
        if a and b:
            graph.setdefault(a, []).append((b, seg))
            graph.setdefault(b, []).append((a, seg))

    # === Step 4: Propagate coordinates ===
    node_coordinates = {origin_node: (0.0, 0.0)}
    queue = deque([origin_node])

    while queue:
        current = queue.popleft()
        current_x, current_y = node_coordinates[current]

        for neighbor, segment in graph.get(current, []):
            if neighbor in node_coordinates:
                continue

            try:
                angle_deg = float(segment.get("absolute_rotation", 0))
                length = float(segment.get("length", 0))
            except ValueError:
                continue

            # Flip the direction if we're traversing from the B-end toward A-end
            if current == segment.get("node_at_end_b"):
                angle_deg = (angle_deg + 180) % 360

            radians = math.radians(angle_deg)
            dx = length * math.cos(radians)
            dy = length * math.sin(radians)

            new_x = round(current_x + dx, 2)
            new_y = round(current_y + dy, 2)
            node_coordinates[neighbor] = (new_x, new_y)
            queue.append(neighbor)

    # === Step 5: Compute and assign average node angles ===
    for node in nodes:
        node_name = node.get("instance_name")
        sum_x, sum_y = 0.0, 0.0
        count = 0

        for seg in segments:
            if (
                seg.get("node_at_end_a") == node_name
                or seg.get("node_at_end_b") == node_name
            ):
                angle_to_add_raw = seg.get("absolute_rotation", "")
                if not angle_to_add_raw:
                    continue
                angle_to_add = float(angle_to_add_raw)

                # Flip 180° if node is at segment_end_b
                if seg.get("node_at_end_b") == node_name:
                    angle_to_add = (angle_to_add + 180) % 360

                # Convert degrees → radians for trig functions
                angle_rad = math.radians(angle_to_add)
                sum_x += math.cos(angle_rad)
                sum_y += math.sin(angle_rad)
                count += 1

        if count:
            # Flip 180° (connector points away from average cable vector)
            avg_x, avg_y = -sum_x, -sum_y

            # Compute angle in degrees, normalized to [0, 360)
            average_angle = math.degrees(math.atan2(avg_y, avg_x)) % 360

            # Round to nearest 0.01 degree
            average_angle = round(average_angle, 2)
        else:
            average_angle = ""

        translate_x, translate_y = node_coordinates.get(node_name, ("", ""))

        instances_list.modify(
            node_name,
            {
                "translate_x": str(translate_x),
                "translate_y": str(translate_y),
                "absolute_rotation": average_angle,
                "parent_csys_instance_name": "origin",
                "parent_csys_outputcsys_name": "origin",
            },
        )

    # === Step 6: Generate PNG ===
    padding = 50
    scale = 96  # pixels per inch
    radius = 5

    # Compute bounding box
    xs = [x for x, y in node_coordinates.values()]
    ys = [y for x, y in node_coordinates.values()]
    min_x, max_x = min(xs), max(xs)
    min_y, max_y = min(ys), max(ys)

    width = int((max_x - min_x) * scale + 2 * padding)
    height = int((max_y - min_y) * scale + 2 * padding)

    def map_xy(x, y):
        """Map logical (CAD) coordinates to image coordinates."""
        return (
            int((x - min_x) * scale + padding),
            int(height - ((y - min_y) * scale + padding)),  # flip Y for image
        )

    # Create white canvas
    img = Image.new("RGB", (width, height), "white")
    draw = ImageDraw.Draw(img)

    # Try to get a system font
    try:
        font = ImageFont.truetype("Arial.ttf", 12)
    except OSError:
        font = ImageFont.load_default()

    # --- Draw segments ---
    for seg in segments:
        a, b = seg.get("node_at_end_a"), seg.get("node_at_end_b")
        if a in node_coordinates and b in node_coordinates:
            x1, y1 = map_xy(*node_coordinates[a])
            x2, y2 = map_xy(*node_coordinates[b])

            # Draw line from A to B
            draw.line((x1, y1, x2, y2), fill="black", width=2)

            # --- Draw arrowhead on B side ---
            arrow_length = 25
            arrow_angle = math.radians(25)  # degrees between arrow sides
            angle = math.atan2(y2 - y1, x2 - x1)

            # Compute arrowhead points
            left_x = x2 - arrow_length * math.cos(angle - arrow_angle)
            left_y = y2 - arrow_length * math.sin(angle - arrow_angle)
            right_x = x2 - arrow_length * math.cos(angle + arrow_angle)
            right_y = y2 - arrow_length * math.sin(angle + arrow_angle)

            draw.polygon([(x2, y2), (left_x, left_y), (right_x, right_y)], fill="black")

            # Midpoint label
            mid_x, mid_y = (x1 + x2) / 2, (y1 + y2) / 2
            draw.text(
                (mid_x, mid_y - 10),
                seg.get("instance_name", ""),
                fill="blue",
                font=font,
            )

    # --- Draw nodes ---
    for name, (x, y) in node_coordinates.items():
        cx, cy = map_xy(x, y)
        draw.ellipse((cx - radius, cy - radius, cx + radius, cy + radius), fill="red")
        draw.text((cx, cy - 15), name, fill="black", font=font, anchor="mm")

    # Legend
    draw.text(
        (padding, height - padding / 2),
        "Arrows point from End A to End B",
        fill="black",
        font=font,
    )

    img.save(fileio.path("formboard graph definition png"), dpi=(96, 96))


def map_instance_to_segments(instance):
    """
    Maps a segment-based instance across multiple segments using pathfinding.

    Takes an instance that spans between two nodes and creates segment-specific
    instances for each segment in the path between those nodes. Uses breadth-first
    search to find the path through the segment graph, then creates new instances
    for each segment with the appropriate direction and order.

    **Note:** The function actually maps to nodes in the same connector group as the
    "end nodes". If your to/from nodes are `item_type=="Cavity"`, for example, this
    function will return paths of segments between the `item_type=="node"` instances
    where those cavities are located.

    **Args:**

    - `instance` (dict): Instance dictionary with `location_type="segment"` that has
        `node_at_end_a` and `node_at_end_b` defined.

    **Raises:**

    - `ValueError`: If the instance is not segment-based, if endpoints are missing,
        if endpoints are not nodes, or if no path is found between the nodes.
    """
    # note to user: we're actually mapping to nodes in same connector group as the "end nodes".
    # so if your to/from nodes are item_type==Cavity, for example, this function will return paths of segments
    # between the item_type==node instance where those cavities are

    # Ensure you're trying to map an instance that is segment-based.
    if instance.get("location_type") != "segment":
        raise ValueError(
            f"You're trying to map a non segment-based instance {instance.get('instance_name')} across segments."
        )

    # Ensure instance has a start and end node
    if instance.get("node_at_end_a") is None or instance.get("node_at_end_b") is None:
        raise ValueError(
            f"Instance {instance.get('instance_name')} has no start or end node."
        )

    # Ensure each endpoint is actually location_type==node
    if (
        instances_list.attribute_of(instance.get("node_at_end_a"), "location_type")
        != "node"
    ):
        raise ValueError(
            f"While mapping '{instance.get('instance_name')}' to segments, location type of {instance.get('node_at_end_a')} is not a node."
        )
    if (
        instances_list.attribute_of(instance.get("node_at_end_b"), "location_type")
        != "node"
    ):
        raise ValueError(
            f"While mapping '{instance.get('instance_name')}' to segments, location type of {instance.get('node_at_end_b')} is not a node."
        )

    # Resolve the node (item_type=="node") for each end's connector group
    start_node_obj = instances_list.instance_in_connector_group_with_item_type(
        instances_list.attribute_of(instance.get("node_at_end_a"), "connector_group"),
        "node",
    )
    try:
        if start_node_obj == 0:
            raise ValueError(
                f"No 'node' type item found in connector group {instance.get('connector_group')}"
            )
        if start_node_obj > 1:
            raise ValueError(
                f"Multiple 'node' type items found in connector group {instance.get('connector_group')}"
            )
    except TypeError:
        pass

    end_node_obj = instances_list.instance_in_connector_group_with_item_type(
        instances_list.attribute_of(instance.get("node_at_end_b"), "connector_group"),
        "node",
    )
    try:
        if end_node_obj == 0:
            raise ValueError(
                f"No 'node' type item found in connector group {instance.get('connector_group')}"
            )
        if end_node_obj > 1:
            raise ValueError(
                f"Multiple 'node' type items found in connector group {instance.get('connector_group')}"
            )
    except TypeError:
        pass

    # Build graph of segments
    segments = [
        inst
        for inst in fileio.read_tsv("instances list")
        if inst.get("item_type") == "segment"
    ]

    graph = {}
    segment_lookup = {}  # frozenset({A, B}) -> seg_name
    seg_endpoints = {}  # seg_name -> (A, B) with stored orientation

    for seg in segments:
        a = seg.get("node_at_end_a")
        b = seg.get("node_at_end_b")
        seg_name = seg.get("instance_name")
        if not a or not b:
            continue
        graph.setdefault(a, set()).add(b)
        graph.setdefault(b, set()).add(a)
        segment_lookup[frozenset([a, b])] = seg_name
        seg_endpoints[seg_name] = (a, b)

    # Re-fetch start/end nodes as instance_names
    start_node_obj = instances_list.instance_in_connector_group_with_item_type(
        instances_list.attribute_of(instance.get("node_at_end_a"), "connector_group"),
        "node",
    )
    if start_node_obj == 0:
        raise ValueError(
            f"No 'node' type item found in connector group {instances_list.attribute_of(instance.get('instance_name'), 'connector_group')}"
        )

    end_node_obj = instances_list.instance_in_connector_group_with_item_type(
        instances_list.attribute_of(instance.get("node_at_end_b"), "connector_group"),
        "node",
    )
    if end_node_obj == 0:
        raise ValueError(
            f"No 'node' type item found in connector group {instances_list.attribute_of(instance.get('instance_name'), 'connector_group')}"
        )

    start_node = start_node_obj.get("instance_name")
    end_node = end_node_obj.get("instance_name")

    # ---- BFS that records per-edge segment + direction ----
    # pred[node] = (prev_node, seg_name, direction)
    pred = {}
    visited = set()
    queue = deque([start_node])
    visited.add(start_node)

    while queue:
        u = queue.popleft()
        if u == end_node:
            break
        for v in graph.get(u, []):
            if v in visited:
                continue
            seg_name = segment_lookup.get(frozenset([u, v]))
            if not seg_name:
                continue
            a_end, b_end = seg_endpoints[seg_name]
            direction = "ab" if (u == a_end and v == b_end) else "ba"
            pred[v] = (u, seg_name, direction)
            visited.add(v)
            queue.append(v)

    if end_node not in pred and start_node != end_node:
        raise ValueError(f"No segment path found between {start_node} and {end_node}")

    # Reconstruct (seg_name, direction) list from predecessors
    segment_steps = []
    if start_node != end_node:
        cur = end_node
        while cur != start_node:
            prev, seg_name, direction = pred[cur]
            segment_steps.append((seg_name, direction))
            cur = prev
        segment_steps.reverse()
    # -------------------------------------------------------

    # Add a new instance for each connected segment, preserving per-segment direction
    i = 1
    for seg_name, direction in segment_steps:
        instances_list.new_instance(
            f"{instance.get('instance_name')}.{seg_name}",
            {
                "item_type": f"{instance.get('item_type')}-segment",
                "parent_instance": instance.get("instance_name"),
                "segment_group": seg_name,
                "segment_order": f"{i}-{direction}",
                "parent_csys": seg_name,
                "location_type": "segment",
                "channel_group": instance.get("channel_group"),
                "circuit_id": instance.get("circuit_id"),
                "circuit_port_number": instance.get("circuit_port_number"),
                "cable_group": instance.get("cable_group"),
                "cable_container": instance.get("cable_container"),
                "cable_identifier": instance.get("cable_identifier"),
                "appearance": instance.get("appearance"),
                "this_net_from_device_refdes": instance.get(
                    "this_net_from_device_refdes"
                ),
                "this_net_from_device_channel_id": instance.get(
                    "this_net_from_device_channel_id"
                ),
                "this_net_from_device_connector_name": instance.get(
                    "this_net_from_device_connector_name"
                ),
                "this_net_to_device_refdes": instance.get("this_net_to_device_refdes"),
                "this_net_to_device_channel_id": instance.get(
                    "this_net_to_device_channel_id"
                ),
                "this_net_to_device_connector_name": instance.get(
                    "this_net_to_device_connector_name"
                ),
                "this_channel_from_device_refdes": instance.get(
                    "this_channel_from_device_refdes"
                ),
                "this_channel_from_device_channel_id": instance.get(
                    "this_channel_from_device_channel_id"
                ),
                "this_channel_to_device_refdes": instance.get(
                    "this_channel_to_device_refdes"
                ),
                "this_channel_to_device_channel_id": instance.get(
                    "this_channel_to_device_channel_id"
                ),
                "this_channel_from_channel_type": instance.get(
                    "this_channel_from_channel_type"
                ),
                "this_channel_to_channel_type": instance.get(
                    "this_channel_to_channel_type"
                ),
                "signal_of_channel_type": instance.get("signal_of_channel_type"),
                "length": instances_list.attribute_of(seg_name, "length"),
            },
        )
        i += 1


def calculate_location(lookup_instance, instances):
    """
    Calculates world coordinates for an instance by accumulating transforms through the CSYS chain.

    Traces the coordinate system hierarchy from the instance up to the origin, accumulating
    translations and rotations at each level. Applies child coordinate system transforms,
    instance translations, and rotations to compute the final world position and angle.

    The function handles both Cartesian (`x`, `y`) and polar (`distance`, `angle`) coordinate
    specifications for child coordinate systems. Absolute rotation overrides accumulated rotation.

    **Args:**

    - `lookup_instance` (dict): The instance dictionary to calculate coordinates for.
    - `instances` (list): List of all instance dictionaries needed to resolve the parent chain.

    **Returns:**

    - `tuple`: A tuple of `(x_pos, y_pos, angle)` representing the world coordinates and
        rotation angle in degrees.

    **Raises:**

    - `ValueError`: If parent coordinate system information is missing or invalid, or
        if parent instances cannot be found in the instances list.
    """
    # ------------------------------------------------------------------
    # Normalize csys_children: ensure all rows contain real dicts
    # ------------------------------------------------------------------
    for instance in instances:
        raw_children = instance.get("csys_children")
        try:
            if isinstance(raw_children, str) and raw_children:
                csys_children = ast.literal_eval(raw_children)
            else:
                csys_children = raw_children or {}
        except Exception:
            csys_children = {}
        instance["csys_children"] = csys_children

    # ------------------------------------------------------------------
    # Build the CSYS parent chain (from item → origin)
    # ------------------------------------------------------------------
    chain = []

    current = lookup_instance

    while True:
        # Find the current instance in the raw instances list
        parent_csys_instance_name = None
        parent_csys_outputcsys_name = None

        for instance in instances:
            if instance.get("instance_name") == current.get("instance_name"):
                parent_csys_instance_name = instance.get("parent_csys_instance_name")
                parent_csys_outputcsys_name = instance.get(
                    "parent_csys_outputcsys_name"
                )
                chain.append(instance)
                break

        # Stop once we reach the origin csys
        if current.get("instance_name") == "origin":
            break

        # Required-parent validation
        if parent_csys_instance_name is None:
            raise ValueError(
                f"Instance '{current.get('instance_name')}' missing parent_csys_instance_name"
            )
        if not parent_csys_instance_name:
            raise ValueError(
                f"Instance '{current.get('instance_name')}' parent_csys_instance_name blank"
            )
        if not parent_csys_outputcsys_name:
            raise ValueError(
                f"Instance '{current.get('instance_name')}' parent_csys_outputcsys_name blank"
            )

        # Resolve parent instance
        parent_csys_instance = None
        for instance in instances:
            if instance.get("instance_name") == parent_csys_instance_name:
                parent_csys_instance = instance
                break
        else:
            raise ValueError(
                f"Instance '{parent_csys_instance_name}' not found in instances list"
            )

        current = parent_csys_instance

    # Reverse chain: now runs origin → lookup_item
    chain = list(reversed(chain))

    # ------------------------------------------------------------------
    # Accumulate transforms along chain
    # ------------------------------------------------------------------
    x_pos = 0.0
    y_pos = 0.0
    angle = 0.0

    for chainlink in chain:
        # ==================================================================
        #   Resolve CHILD CSYS (the transform from parent output csys)
        # ==================================================================
        relevant_csys_child = None

        # Find the chainlink's parent instance and extract its csys_children
        for instance in instances:
            if instance.get("instance_name") == chainlink.get(
                "parent_csys_instance_name"
            ):
                if instance.get("csys_children") != {}:
                    relevant_csys_child = instance["csys_children"].get(
                        chainlink.get("parent_csys_outputcsys_name")
                    )
                else:
                    relevant_csys_child = {}
                break

        angle_old = angle
        dx = 0.0
        dy = 0.0

        # ------------------------------------------------------------------
        # Child CSYS: translation component
        # ------------------------------------------------------------------
        if relevant_csys_child is not None:
            # x/y explicit translation
            if relevant_csys_child.get("x") not in [
                "",
                None,
            ] and relevant_csys_child.get("y") not in ["", None]:
                dx = float(relevant_csys_child.get("x"))
                dy = float(relevant_csys_child.get("y"))

            # polar translation
            elif relevant_csys_child.get("distance") not in [
                "",
                None,
            ] and relevant_csys_child.get("angle") not in ["", None]:
                dist = float(relevant_csys_child.get("distance"))
                ang = math.radians(float(relevant_csys_child.get("angle")))
                dx = dist * math.cos(ang)
                dy = dist * math.sin(ang)

            # Child CSYS rotation
            if relevant_csys_child.get("rotation") not in ["", None]:
                angle += float(relevant_csys_child.get("rotation"))

        # ------------------------------------------------------------------
        # Apply rotated dx/dy into world coordinates
        # ------------------------------------------------------------------
        dx_old = dx
        dy_old = dy

        x_pos += dx_old * math.cos(math.radians(angle_old)) - dy_old * math.sin(
            math.radians(angle_old)
        )
        y_pos += dx_old * math.sin(math.radians(angle_old)) + dy_old * math.cos(
            math.radians(angle_old)
        )

        # ------------------------------------------------------------------
        # Chainlink's local transform fields
        # ------------------------------------------------------------------
        if chainlink.get("translate_x") not in ["", None]:
            x_pos += float(chainlink.get("translate_x"))

        if chainlink.get("translate_y") not in ["", None]:
            y_pos += float(chainlink.get("translate_y"))

        if chainlink.get("rotate_csys") not in ["", None]:
            angle += float(chainlink.get("rotate_csys"))

        # Absolute rotation overrides accumulated rotation
        if chainlink.get("absolute_rotation") not in ["", None]:
            angle = float(chainlink.get("absolute_rotation"))

    # ------------------------------------------------------------------
    # Final world coordinates
    # ------------------------------------------------------------------
    return x_pos, y_pos, angle


def draw_line(
    from_coords,
    to_coords,
    scale=1,
    from_leader=False,
    to_leader=True,
    indent=6,
    stroke="black",
    thickness=1,
):
    """
    Generates SVG markup for a line with optional arrowheads.

    Creates SVG elements for a line connecting two points, with optional arrowheads
    at either or both ends. Coordinates are converted from inches to pixels (96 dpi)
    with Y-axis flipped for SVG coordinate system.

    **Args:**
    - `from_coords` (tuple): Starting coordinates as `(x, y)` in inches.
    - `to_coords` (tuple): Ending coordinates as `(x, y)` in inches.
    - `scale` (float, optional): Scale factor for arrowhead size and line thickness. Defaults to `1`.
    - `from_leader` (bool, optional): If `True`, draw an arrowhead at the start. Defaults to `False`.
    - `to_leader` (bool, optional): If `True`, draw an arrowhead at the end. Defaults to `True`.
    - `indent` (int, optional): Unused parameter (maintained for compatibility). Defaults to `6`.
    - `stroke` (str, optional): Stroke color. Defaults to `"black"`.
    - `thickness` (float, optional): Line thickness in pixels (before scaling). Defaults to `1`.

    **Returns:**
    - `str`: SVG markup string containing the line and arrowhead elements, or empty
        string if coordinates are identical (zero-length line).
    """
    # Convert inches → px (and flip Y)
    fx, fy = from_coords[0] * 96, from_coords[1] * -96
    tx, ty = to_coords[0] * 96, to_coords[1] * -96

    # Geometry for arrowheads
    dx = tx - fx
    dy = ty - fy
    line_len = math.hypot(dx, dy)

    if line_len == 0:
        return ""  # cannot draw arrowheads on zero-length line

    # Normalize direction vector
    ux = dx / line_len
    uy = dy / line_len

    # Arrowhead size (scaled)
    arrow_len = 8 / scale
    arrow_wid = 6 / scale

    svg_parts = []

    # -------------------------
    # Draw main line
    # -------------------------
    svg_parts.append(
        f'<line x1="{fx}" y1="{fy}" x2="{tx}" y2="{ty}" '
        f'stroke="{stroke}" stroke-width="{thickness / scale}"/>'
    )

    # -------------------------
    # Arrowhead helper
    # -------------------------
    def arrowhead(px, py, ux, uy):
        # Base of arrowhead is at (px, py)
        # Two side points:
        # Rotate ±90° for perpendicular direction
        perp_x = -uy
        perp_y = ux

        p1x = px - ux * arrow_len + perp_x * (arrow_wid / 2)
        p1y = py - uy * arrow_len + perp_y * (arrow_wid / 2)

        p2x = px - ux * arrow_len - perp_x * (arrow_wid / 2)
        p2y = py - uy * arrow_len - perp_y * (arrow_wid / 2)

        return f'<polygon points="{px},{py} {p1x},{p1y} {p2x},{p2y}" fill="{stroke}"/>'

    # -------------------------
    # Arrow at TO end
    # -------------------------
    if to_leader:
        svg_parts.append(arrowhead(tx, ty, ux, uy))

    # -------------------------
    # Arrow at FROM end
    # reverse direction for correct orientation
    # -------------------------
    if from_leader:
        svg_parts.append(arrowhead(fx, fy, -ux, -uy))

    return "".join(svg_parts)



import csv
import os
from harnice import fileio

list_type = None
COLUMNS = []

# Signals list column headers to match source of truth + compatibility change
DEVICE_COLUMNS = [
    "channel_id",  # Unique identifier for the channel.
    "signal", # Name of the electrical function of that signal, as it pertains to its channel type defition. i.e. "positive"
    "connector_name", # Unique identifier for the connector that this signal and channel is a part of.
    "cavity", # Identifier of the pin, socket, stud, etc, that this signal is internally electrically routed to within its connector.
    "connector_mpn", #MPN of the connector in this device (NOT the mating connector).
    "channel_type",  #The channel type of this signal. \n{% include-markdown "fragments/channel_type_reference.md" %}
    "config_variable", #Change header or add more headers as needed. Blank: row is true across all values of this field. Otherwise, row is only true when configuration matches the value of this field.
]

DISCONNECT_COLUMNS = [
    "channel_id",  # Unique identifier for the channel.
    "signal",# Name of the electrical function of that signal, as it pertains to its channel type defition. i.e. "positive"
    "A_cavity", #Identifier of the pin, socket, stud, etc, that this signal is internally electrically routed to within that side of the connector.\n??? question "Why are A and B different here?"\n    Sometimes it's possible to have connectors that have cavities that may mate electrically, but have different names. For example, suppose two connectors physically mate, but are made by different manufacturers. One manufacturer used lowercase (a, b, c) to reference the cavities but the other used uppercase (A, B, C), or numbers (1, 2, 3), or colors (red, green, blue), etc.
    "B_cavity" #Identifier of the pin, socket, stud, etc, that this signal is internally electrically routed to within that side of the connector.\n??? question "Why are A and B different here?"\n    Sometimes it's possible to have connectors that have cavities that may mate electrically, but have different names. For example, suppose two connectors physically mate, but are made by different manufacturers. One manufacturer used lowercase (a, b, c) to reference the cavities but the other used uppercase (A, B, C), or numbers (1, 2, 3), or colors (red, green, blue), etc.
    "A_connector_mpn", #MPN of the connector of the harness on this side of the disconnect
    "A_channel_type", #The channel type of this side of the discconect.\n??? question "Why are A and B different here?"\n    It's important to keep track of which side has which channel type so that you cannot accidentally flip pins and sockets, for example, by mapping the wrong channel type to the wrong pin gender. Careful validation should be done when mapping channels through disconnects to ensure the disconnects have channels that pass through them in the correct direction.
    "B_connector_mpn",#MPN of the connector of the harness on this side of the disconnect
    "B_channel_type", #The channel type of this side of the discconect.\n??? question "Why are A and B different here?"\n    It's important to keep track of which side has which channel type so that you cannot accidentally flip pins and sockets, for example, by mapping the wrong channel type to the wrong pin gender. Careful validation should be done when mapping channels through disconnects to ensure the disconnects have channels that pass through them in the correct direction.
]


def set_list_type(x):
    global list_type
    list_type = x

    global COLUMNS
    if list_type == "device":
        COLUMNS = DEVICE_COLUMNS
    elif list_type == "disconnect":
        COLUMNS = DISCONNECT_COLUMNS


def new():
    """
    Creates a new signals TSV file at fileio.path("signals list") with only the header row.
    Overwrites any existing file.
    """
    signals_path = fileio.path("signals list")
    os.makedirs(os.path.dirname(signals_path), exist_ok=True)

    if os.path.exists(signals_path):
        os.remove(signals_path)

    with open(signals_path, "w", newline="", encoding="utf-8") as f:
        writer = csv.writer(f, delimiter="\t")
        writer.writerow(COLUMNS)


def append(**kwargs):
    """
    Appends a new row to the signals TSV file.
    Missing optional fields will be written as empty strings.
    Raises ValueError if required fields are missing.

    Required kwargs:
        For 'device':
            channel_id, signal, connector_name, cavity, connector_mpn, channel_type
        For 'disconnect':
            A_channel_id, A_signal, A_connector_name, A_cavity, A_connector_mpn, A_channel_type,
            B_channel_id, B_signal, B_connector_name, B_cavity, B_connector_mpn, B_channel_type
    """
    signals_path = fileio.path("signals list")

    # Create the signals list file if it doesn't exist
    if not os.path.exists(signals_path):
        new()

    # --- Define required fields based on product type ---
    if list_type == "device":
        required = [
            "channel_id",
            "signal",
            "connector_name",
            "cavity",
            "connector_mpn",
            "channel_type",
        ]
    elif list_type == "disconnect":
        required = [
            "channel_id",
            "signal",
            "A_cavity",
            "A_connector_mpn",
            "A_channel_type",
            "B_cavity",
            "B_connector_mpn",
            "B_channel_type",
        ]
    else:
        required = []

    # --- Check for missing required fields ---
    missing = [key for key in required if not kwargs.get(key)]
    if missing:
        raise ValueError(
            f"Missing required signal fields for '{list_type}': {', '.join(missing)}"
        )

    # --- Fill row in header order ---
    row = [kwargs.get(col, "") for col in COLUMNS]

    # --- Append to the signals list ---
    with open(signals_path, "a", newline="", encoding="utf-8") as f:
        writer = csv.writer(f, delimiter="\t", lineterminator="\n")
        writer.writerow(row)


def cavity_of_signal(channel_id, signal, path_to_signals_list):
    for row in fileio.read_tsv(path_to_signals_list):
        if row.get("signal", "").strip() == signal.strip():
            if row.get("channel_id", "").strip() == channel_id.strip():
                return row.get("cavity", "").strip()
    raise ValueError(
        f"Signal {signal} of channel_id {channel_id} not found in {path_to_signals_list}"
    )


def connector_name_of_channel(channel_id, path_to_signals_list):
    if not os.path.exists(path_to_signals_list):
        raise FileNotFoundError(f"Signals list file not found: {path_to_signals_list}")

    with open(path_to_signals_list, newline="", encoding="utf-8") as f:
        reader = csv.DictReader(f, delimiter="\t")
        for row in reader:
            if row.get("channel_id", "").strip() == channel_id.strip():
                return row.get("connector_name", "").strip()
